from abc import ABC, abstractmethod
from typing import Dict

import fasttext
import pandas as pd

from mergenetic import PROJECT_ROOT

# ==========================
#  BASE EVALUATOR CLASSES
# ==========================


class BaseEvaluator(ABC):
    """
    Abstract base class for all evaluators.
    """

    data: pd.DataFrame

    @abstractmethod
    def get_correctness(self, dataframe: pd.DataFrame, **kwargs):
        """
        Abstract method to return computed correctness.
        """
        pass

    @abstractmethod
    def get_data(self) -> pd.DataFrame:
        """
        Abstract method to return the predictions DataFrame.
        """
        pass

    @staticmethod
    def _validate_dataframe(df: pd.DataFrame) -> None:
        """Ensures that the DataFrame has the required columns."""
        if not isinstance(df, pd.DataFrame):
            raise TypeError("Input must be a pandas DataFrame.")
        if "predictions" not in df.columns or "answer" not in df.columns:
            raise ValueError(
                "DataFrame must contain 'predictions' and 'answer' columns."
            )


class BaseMultiObjectiveEvaluator(BaseEvaluator):
    """
    Abstract base class for multi-objective evaluators.
    """

    data: Dict[str, pd.DataFrame]

    @abstractmethod
    def get_correctness(self, dataframes_dict: Dict[str, pd.DataFrame], **kwargs):
        """
        Abstract method to return computed correctness.
        """
        pass

    @abstractmethod
    def get_data(self) -> Dict[str, pd.DataFrame]:
        """
        Abstract method to return a dictionary of predictions DataFrames.
        """
        pass


# =====================
#  LANGUAGE DETECTOR
# =====================
class LanguageDetector:
    """
    Detects the language of a given text using FastText.
    """

    SUPPORTED_LANGUAGES = set(
        """
        af als am an ar arz as ast av az azb ba bar bcl be bg bh bn bo bpy br bs bxr ca cbk ce ceb ckb co cs cv 
        cy da de diq dsb dty dv el eml en eo es et eu fa fi fr frr fy ga gd gl gn gom gu gv he hi hif hr hsb ht hu 
        hy ia id ie ilo io is it ja jbo jv ka kk km kn ko krc ku kv kw ky la lb lez li lmo lo lrc lt lv mai mg mhr 
        min mk ml mn mr mrj ms mt mwl my myv mzn nah nap nds ne new nl nn no oc or os pa pam pfl pl pms pnb ps pt 
        qu rm ro ru rue sa sah sc scn sco sd sh si sk sl so sq sr su sv sw ta te tg th tk tl tr tt tyv ug uk ur uz 
        vec vep vi vls vo wa war wuu xal xmf yi yo yue zh
    """.split()
    )

    def __init__(
        self,
        langs: list[str],
        fasttext_model_path: str = f"{PROJECT_ROOT}/models/fasttext/lid.176.bin",
    ):
        """
        Initializer for the language detector.

        Parameters
        ----------
        langs: list[str]
            List with languages used for language detection.
        fasttext_model_path: str
            Path to the FastText model.
        """
        self.detector = fasttext.load_model(fasttext_model_path)

        unsupported_languages = [
            lang for lang in langs if lang not in self.SUPPORTED_LANGUAGES
        ]
        if unsupported_languages:
            raise ValueError(
                f"Unsupported languages detected: {unsupported_languages}."
            )

    def _get_language(self, text: str) -> str:
        """
        Detects the language of a given text using FastText.

        Parameters
        ----------
        text : str
            The text to detect the language from.

        Returns
        -------
        str
            Detected language label or "UNK" if confidence is too low.
        """
        text_clean = text.replace("\n", " ")
        pred_lang, confidence = self.detector.predict(text_clean)

        if confidence[0] < 0.3:
            return "UNK"
        elif confidence[0] < 0.5:
            return self.detector.predict(" ".join(text_clean.split()[1:]))[0]
        return pred_lang[0] if confidence[0] >= 0.5 else "UNK"

    def get_supported_languages(self) -> set[str]:
        """
        Returns the set of supported languages.

        Returns
        -------
        set[str]
            Set of supported languages.
        """
        return self.SUPPORTED_LANGUAGES



import logging
from pathlib import Path
from typing import Iterable

import yaml

from mergenetic.merging.merger import Merger

logger = logging.getLogger(__name__)


class LinearMerger(Merger):
    def __init__(
        self,
        run_id: str,
        path_to_base_model: str,
        model_paths: list,
        path_to_store_yaml: str,
        path_to_store_merged_model: str,
        dtype: str,
    ) -> None:
        """
        Concrete class for merging models. It is used to create a configuration file for merging models according to Linear.
        It requires all the information for merging through mergekit library according to the given merging technique.
        Two type of information are used:
            1. Static information common to all merged models with that technique (e.g., base model, dtype, etc.); these are passed through the class init.
            2. Dynamic information specific to each merging operation (e.g., weights, densities, etc.); these are passed through the method for merging.

        Parameters
        ----------
        run_id : str
            is the id of the run that will be used to store all the yaml configuration files for merging.
        path_to_base_model : str
            is the path to the downloaded base model that will be used for merging.
        model_paths : list
            variable number of paths for the models that will be merged.
        path_to_store_yaml : str
            is the path to the folder where the yaml configuration files for merging will be stored.
        path_to_store_merged_model : str
            is the path to the folder where the merged models will be stored.
        dtype : str
            is the data type of the models (e.g., float16, float32, etc.).

        Returns
        -------
        None
        """
        super().__init__(run_id, path_to_store_yaml, path_to_store_merged_model, dtype)

        # store the paths to the models for ties dare
        self.path_to_base_model = Path(path_to_base_model)
        self.model_paths = [Path(path) for path in model_paths]

    def create_individual_configuration(self, weights: Iterable) -> Path | str:
        """
        Method to create a configuration file for merging models according to DARE_TIES mergekit implementation.
        It requires dynamic information specific to each merging operation (e.g., weights, densities, etc.).
        It creates the configuration file based on this info and class attributes.
        Store the configuration file in the folder specified in the class attributes.

        Parameters
        ----------
        weights_and_densities : Iterable
            list of pairs of weights and densities for each model.

        Returns
        -------
        Path or str
            is the path to the yaml configuration file created.
        """
        # Create a list of dictionaries for the models and their parameters
        weights = [float(weight) for weight in weights]

        # Create a dictionary for the configuration file
        # model_info = [{'model': str(self.path_to_base_model)}]
        model_info = []
        for model_path, weight in zip(self.model_paths, weights):
            model_info.append(
                {"model": str(model_path), "parameters": {"weight": weight}}
            )

        config = {
            #'tokenizer_source': 'base', # TODO: this is used to solve mismatching of tokenizers, but it is not the best solution to hardcode it
            "models": model_info,
            "merge_method": "linear",
            #'base_model': str(self.path_to_base_model),
            "parameters": {"int8_mask": True},
            "dtype": self.dtype,
        }

        # Create Directory for YAML file
        config_directory = Path(self.path_to_store_yaml).parent
        config_directory.mkdir(parents=True, exist_ok=True)

        # Write the dictionary to a YAML file
        with open(self.path_to_store_yaml, "w") as file:
            yaml.dump(config, file, default_flow_style=False)
            logger.info(f"Configuration file created at {self.path_to_store_yaml}")

        # Return the path to the YAML file
        return Path(self.path_to_store_yaml)



import re
from logging import getLogger
from typing import Dict, List, Tuple, Union

import pandas as pd

from .evaluator import BaseMultiObjectiveEvaluator, LanguageDetector

logger = getLogger(__name__)

# =========================================
# MULTILINGUAL MULTIPLE CHOICE EVALUATOR
# =========================================


class MultilingualMCEvaluator(BaseMultiObjectiveEvaluator):
    """
    Base class for multiple-choice question evaluators, providing common methods for
    extracting answers for multiple-choice questions, detecting languages, and computing accuracy.
    """

    lang_detector: LanguageDetector

    def __init__(self, language_ids: list[str] = None, validate_lang=True) -> None:
        """
        Initializes the evaluator with a dictionary of dataframes, each corresponding to a language.

        Parameters
        ---------
        language_ids : list[str]
            List of language IDs for Language Detection.
        validate_lang : bool
            If True, validates the language of the predictions.
        """

        self.language_ids = language_ids
        self.validate_lang = validate_lang

        if validate_lang:
            try:
                self.lang_detector = LanguageDetector(language_ids)
            except Exception as e:
                logger.warning(f"Language detection disabled: {e}")
                self.lang_detector = None
        else:
            self.lang_detector = None

    @staticmethod
    def _extract_answer(text: str) -> str:
        """
        Extracts the selected option from the model's prediction.

        Parameters
        ----------
        text : str
            The text from which to extract the answer.

        Returns
        -------
        str
            The extracted answer option (e.g., 'A', 'B', 'C', 'D'), or 'None' if no match is found.
        """
        matches = re.findall(r"\(?([A-Z])\)", text)
        return matches[-1] if matches else "None"

    def get_correctness(
        self, dataframe_dict: Dict[str, pd.DataFrame]
    ) -> Union[float, Tuple[List[float], str]]:
        """
        Computes correctness across multiple language datasets.

        Parameters
        ----------
        dataframe_dict : dict[str, pd.DataFrame]
            Dictionary of dataframes, each corresponding to a language.

        Returns
        -------
        float | tuple[list[float], str]
            Per-language correctness.
        """
        self.data: dict[str, pd.DataFrame] = {}
        for k, v in dataframe_dict.items():
            self._validate_dataframe(v)
            self.data[k] = v.copy()

        for lang, df in self.data.items():
            df["predictions_filtered"] = df["predictions"].apply(self._extract_answer)
            df["language"] = df["predictions"].apply(self.lang_detector._get_language)

        def calc_correctness(df: pd.DataFrame, lang: str = None):
            if self.validate_lang:
                if self.lang_detector is None:
                    raise ValueError(
                        "Language detector must be provided for language validation."
                    )

                correctness = (df["predictions_filtered"] == df["answer"]) & (
                    (df["language"] == f"__label__{lang}") | (df["language"] == "UNK")
                )

            else:
                correctness = df["predictions_filtered"] == df["answer"]

            df["is_correct"] = correctness
            return correctness

        correctness_dict: Dict[str, pd.Series] = {}
        for lang, df in self.data.items():
            correctness_dict[lang] = calc_correctness(df, lang)

        return correctness_dict

    def get_data(self) -> Dict[str, pd.DataFrame]:
        """
        Returns the DataFrame with the results.

        Returns
        -------
        dict[str, pd.DataFrame]
            The DataFrame with the results.
        """
        return self.data


# ==============================================
# MULTILINGUAL MATH FREE GENERATION EVALUATOR
# ==============================================
class MultilingualMathFGEvaluator(BaseMultiObjectiveEvaluator):
    """
    Evaluator for multilingual math datasets on model generated answers.
    """

    lang_detector: LanguageDetector

    def __init__(
        self, language_ids: list[str] = [], validate_lang: bool = True
    ) -> None:
        """
        Parameters
        ----------
        fasttext_model_path : str
            Path to the FastText model for language detection.

        """
        self.language_ids = language_ids
        self.validate_lang = validate_lang

        if validate_lang:
            self.lang_detector = LanguageDetector(language_ids)

    @staticmethod
    def _extract_answer(text: str) -> str:
        """
        Extracts the last number from the model's prediction.

        Parameters
        ----------
        text : str
            The text from which to extract the answer.

        Returns
        -------
        str
            The extracted number, or 'None' if no match is found.
        """
        matches = re.findall(r"\d+", text)
        return matches[-1] if matches else "None"

    def get_correctness(
        self,
        dataframes_dict: Dict[str, pd.DataFrame],
    ) -> Dict[str, pd.Series]:
        """
        Computes the correctness of the model for math problems with free generation.

        Parameters
        ----------
        dataframes_dict : dict[str, pd.DataFrame]
            Dictionary of dataframes, each corresponding to a language.

        Returns
        -------
        dict[str, float] | tuple[dict[str, float], dict[str, pd.DataFrame]]
            Per-language correctness.
        """
        correctness_dict = {}

        self.data: Dict[str, pd.DataFrame] = {}
        for k, v in dataframes_dict.items():
            self._validate_dataframe(v)
            self.data[k] = v.copy()

        for df in self.data.values():
            df["predictions_filtered"] = df["predictions"].apply(self._extract_answer)
            df["is_correct"] = df["predictions_filtered"] == df["answer"]

            if self.validate_lang:
                df["language"] = df["predictions"].apply(
                    self.lang_detector._get_language
                )

        def calc_correctness(df: pd.DataFrame, lang: str = None):
            if not self.validate_lang:
                return df["is_correct"]
            else:
                return df["is_correct"] & (
                    (df["language"] == f"__label__{lang}") | (df["language"] == "UNK")
                )

        for lang, df in self.data.items():
            correctness_dict[lang] = calc_correctness(df, lang)

        return correctness_dict

    def get_data(self) -> Dict[str, pd.DataFrame]:
        """
        Returns the DataFrame with the results.

        Returns
        -------
        dict[str, pd.DataFrame]
            The DataFrame with the results.
        """
        return self.data



import logging
from pathlib import Path
from typing import Iterable

import yaml

from mergenetic.merging.merger import Merger

logger = logging.getLogger(__name__)


class SlerpMerger(Merger):
    def __init__(
        self,
        run_id: str,
        path_to_base_model: Path | str,
        layer_range_base_model: str,
        path_to_model_1: Path | str,
        layer_range_model_1: str,
        path_to_store_yaml: Path | str,
        path_to_store_merged_model: Path | str,
        dtype: str,
    ) -> None:
        """
        Concrete class for merging models. It is used to create a configuration file for merging models according to Slerp.
        It requires all the information for merging through mergekit library according to the given merging technique. Two type of information are used:
            1. Static information common to all merged models with that technique (e.g., base model, dtype, etc.); these are passed through the class init.
            2. Dynamic information specific to each merging operation (e.g., t, layer range); these are passed through the method for merging.

        Parameters
        ----------
        run_id : str
            is the id of the run that will be used to store all the yaml configuration files for merging.
        path_to_base_model : str
            is the path to the downloaded base model that will be used for merging.
        layer_range_base_model : list
            is the list of layer indices for the base model that takes part in the interpolation.
        path_to_model_1 : str
            is the path to the first model that will be merged.
        layer_range_model_1 : list
            is the list of layer indices for the first model that takes part in the interpolation.
        path_to_store_yaml : str
            is the path to the folder where the yaml configuration files for merging will be stored.
        path_to_store_merged_model : str
            is the path to the folder where the merged models will be stored.
        dtype : str
            is the data type of the models (e.g., float16, float32, etc.).

        Returns
        -------
        None
        """
        # init merger common attributes
        super().__init__(run_id, path_to_store_yaml, path_to_store_merged_model, dtype)

        # store the paths to the models for ties dare
        self.path_to_base_model = Path(path_to_base_model)
        self.path_to_model_1 = Path(path_to_model_1)
        self.layer_range_base_model = layer_range_base_model
        self.layer_range_model_1 = layer_range_model_1

    def create_individual_configuration(self, x: Iterable) -> Path | str:
        """
        Abstract method to create a configuration file for merging models according to SLERP mergekit implementation.
        It requires dynamic information specific to each merging operation (e.g., t, layer range).
        It creates the configuration file based on this info and class attributes.
        store the configuration file in the folder specified in the class attributes.

        Parameters
        ----------
        x : Iterable
            is the list/array of values used to instatiate the configuration file of mergekit for merging through slerp.

        Returns
        -------
        Path | str
            is the path to the yaml configuration file created.

        """

        # convert the array from pymoo to list of float
        x = [float(i) for i in x]

        # extract the values of t for self_attn, mlp, and other
        t_self_attn = x[0:5]
        t_mlp = x[5:10]
        t_other = x[10]

        # create a configuration file from a dictionary
        config = {
            "slices": [
                {
                    "sources": [
                        {
                            "model": str(self.path_to_base_model),
                            "layer_range": self.layer_range_base_model,
                        },
                        {
                            "model": str(self.path_to_model_1),
                            "layer_range": self.layer_range_model_1,
                        },
                    ]
                }
            ],
            "merge_method": "slerp",
            "base_model": str(self.path_to_base_model),
            "parameters": {
                "t": [
                    {"filter": "self_attn", "value": t_self_attn},
                    {"filter": "mlp", "value": t_mlp},
                    {"value": t_other},
                ]
            },
            "dtype": self.dtype,
        }

        # Create Directory for YAML fil
        config_directory = self.path_to_store_yaml.parent
        config_directory.mkdir(parents=True, exist_ok=True)

        # Write the dictionary to a YAML file
        with open(self.path_to_store_yaml, "w") as file:
            yaml.dump(config, file, default_flow_style=False)
            logger.info(f"Configuration file created at {self.path_to_store_yaml}")

        # Return the path to the YAML file
        return self.path_to_store_yaml



from abc import ABC, abstractmethod
from logging import getLogger
from pathlib import Path
from typing import Dict, Optional

import numpy as np
import pandas as pd
import torch
from lm_eval.models.vllm_causallms import VLLM
from pymoo.core.problem import Problem
from transformers import AutoModelForCausalLM, AutoTokenizer, BitsAndBytesConfig

from mergenetic.merging.merger import Merger

logger = getLogger(__name__)


class BaseMergingProblem(ABC, Problem):
    """
    Base class for merging problems (single and multi-objective).
    """

    def __init__(
        self,
        merger: Merger,
        n_var: int,
        n_obj: int,
        xl=0,
        xu=1,
        n_eq_constr: int = 0,
        n_ieq_constr: int = 0,  # TODO: verify if needed for single-objective
        eval_batch_size: int = 5,
        discrete: bool = False,
        device: str = "cuda",
        load_in_4bit: bool = True,
        use_lm_eval: bool = False,  # Flag for using lm-eval-harness (VLLM)
        verbose_evaluation: bool = True,
        test_mode: bool = False,
        custom_prompt_template=False,
    ) -> None:

        self.n_var = n_var
        self.n_obj = n_obj
        self.merger = merger
        self.discrete = discrete
        self.eval_batch_size = eval_batch_size
        self.device = device
        self.load_in_4bit = load_in_4bit
        self.use_lm_eval = use_lm_eval  # Enables VLLM
        self.verbose_evaluation = verbose_evaluation
        self.test_mode = test_mode
        self.custom_prompt_template = custom_prompt_template or False
        self.n_eq_constr = n_eq_constr
        self.n_ieq_constr = n_ieq_constr

        if self.load_in_4bit:
            logger.info("Loading model in 4bit...")

        # Adjust bounds for discrete search
        if self.discrete:
            xl, xu = 0, 10

        # Initialize pymoo problem
        super().__init__(elementwise=True, n_var=n_var, n_obj=n_obj, xl=xl, xu=xu)

        # Tracking performance
        self.objective_list = [f"objective_{i+1}" for i in range(n_obj)]
        self.phenotype_feature_list = [f"genotype_{i+1}" for i in range(n_var)]

        # Single-objective case: one DataFrame
        if self.n_obj == 1:
            self.results_df = pd.DataFrame(
                columns=self.objective_list + self.phenotype_feature_list + ["step"]
            )
        else:
            # Multi-objective case: multiple fitness DataFrames (one per objective)
            self.results_df = {
                obj: pd.DataFrame(
                    columns=[obj] + self.phenotype_feature_list + ["step"]
                )
                for obj in self.objective_list
            }

        self.step = 0

    def _from_array_to_genotype(self, x: np.array) -> Path:
        """Convert an array representation to a model configuration."""
        logger.info(f"Converting genotype to configuration: {x}")
        path_to_config = self.merger.create_individual_configuration(x)
        return self.merger.merge_model_from_configuration(path_to_config)

    def load_model(
        self, path: str, verbose: bool = True
    ) -> VLLM | tuple[AutoModelForCausalLM, AutoTokenizer]:
        """Loads a model and tokenizer using either Hugging Face or VLLM."""
        logger.info(f"Loading model from: {path}")
        logger.info(f"On device: {self.device}. Load in 4bit: {self.load_in_4bit}")

        device = self.device if self.device is not None else "cpu"
        load_on_gpu = "cuda" in device

        if load_on_gpu and not torch.cuda.is_available():
            raise RuntimeError("CUDA is not available, but GPU loading was requested.")

        model: AutoModelForCausalLM = None

        if self.use_lm_eval:
            return VLLM(
                pretrained=str(path),
                device=device,
                dtype=torch.bfloat16,
                quantization="bitsandbytes" if self.load_in_4bit else None,
                gpu_memory_utilization=0.8 if self.load_in_4bit else 0.9,
            )
        else:
            if self.load_in_4bit:
                quant_config = BitsAndBytesConfig(
                    load_in_4bit=True,
                    bnb_4bit_quant_type="nf4",
                    bnb_4bit_use_double_quant=True,
                    bnb_4bit_compute_dtype=torch.float16,
                )
                model = AutoModelForCausalLM.from_pretrained(
                    path, device_map=device, quantization_config=quant_config
                )
            else:
                model = AutoModelForCausalLM.from_pretrained(path, device_map=device)

        model.config.use_cache = True
        model.eval()
        tokenizer = AutoTokenizer.from_pretrained(path, padding_side="left")

        return model, tokenizer

    @abstractmethod
    def metrics_4_genotype(
        self, model, tokenizer=None
    ) -> tuple[list[float], Optional[str]]:
        """Abstract method for evaluating a model."""
        pass

    @abstractmethod
    def test(self):
        """Abstract method for testing the model."""
        pass

    @abstractmethod
    def get_data(self) -> pd.DataFrame:
        """Abstract method for getting the data."""
        pass

    def _evaluate(self, x, out, *args, **kwargs):
        """Main evaluation function for optimization."""
        if self.discrete:
            x = x / 10  # Normalize discrete values

        path_to_model = self._from_array_to_genotype(x)
        model = self.load_model(path_to_model)

        logger.info(f"Evaluating model at step {self.step}: {x}")
        # Multi-objective fitness handling
        if self.use_lm_eval:
            f, description = self.metrics_4_genotype(model)
        else:
            model, tokenizer = model  # Unpack if not using VLLM
            f, description = self.metrics_4_genotype(model, tokenizer)

        logger.info(f"Model evaluation results: {f}")

        out["F"] = f
        self.step += 1

        log_entry = dict(
            zip(self.phenotype_feature_list, x.flatten()),
            **dict(zip(self.objective_list, f)),
            step=self.step,
        )

        if self.n_obj == 1:
            self.results_df = pd.concat(
                [self.results_df, pd.DataFrame([log_entry])], ignore_index=True
            )
        else:
            # Update each objective's DataFrame separately
            for obj_name, fitness_value in zip(self.objective_list, f):
                self.results_df[obj_name] = pd.concat(
                    [
                        self.results_df[obj_name],
                        pd.DataFrame([{**log_entry, obj_name: fitness_value}]),
                    ],
                    ignore_index=True,
                )

        del model
        torch.cuda.empty_cache()

        if description:
            logger.debug(f"Individual {x} description: {description}")


class MergingProblem(BaseMergingProblem):
    """
    Implementation of a **single-objective** merging problem.
    """

    def __init__(
        self,
        merger,
        search_df: pd.DataFrame,
        test_df: Optional[pd.DataFrame] = None,
        **kwargs,
    ):
        super().__init__(merger, **kwargs)  # Enforce single-objective
        self.search_df: pd.DataFrame = search_df
        self.test_df: Optional[pd.DataFrame] = test_df

    def get_data(self) -> pd.DataFrame:
        return self.results_df


class MultiObjectiveMergingProblem(BaseMergingProblem):
    """
    Implementation of a multi-objective optimization problem for model merging.

    This class extends BaseMergingProblem to handle multiple competing objectives
    simultaneously (e.g., optimizing for both accuracy and efficiency). Unlike
    single-objective optimization, this creates a Pareto front of solutions
    representing different trade-offs between objectives.

    Each objective requires its own evaluation dataset and metrics. Results are
    tracked separately for each objective, allowing for comprehensive analysis
    of the optimization landscape and trade-offs between competing goals.
    """

    def __init__(
        self,
        merger,
        n_obj: int,
        search_dataframes: Dict[str, pd.DataFrame] | None,
        test_dataframes: Optional[Dict[str, pd.DataFrame]],
        **kwargs,
    ):
        assert (
            n_obj > 1
        ), "MultiObjectiveMergingProblem must have more than one objective!"
        super().__init__(merger, n_obj=n_obj, **kwargs)
        self.search_dataframes: Dict[str, pd.DataFrame] | None = search_dataframes
        self.test_dataframes: Optional[Dict[str, pd.DataFrame]] = test_dataframes

    def get_data(self) -> Dict[str, pd.DataFrame]:
        return self.results_df



import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd

from mergenetic.estimator.perf_estimation import (
    PerformanceEstimationParameters,
    PerformanceEstimator,
)
from mergenetic.evaluation.lm_harness import LmHarnessEvaluator
from mergenetic.evaluation.math_language import FGMathEvaluator
from mergenetic.evaluation.multilingual_evaluator import MultilingualMCEvaluator
from mergenetic.optimization import MergingProblem, MultiObjectiveMergingProblem
from mergenetic.utils import get_batched_model_predictions

logger = logging.getLogger(__name__)


@dataclass
class ConfigPE:
    thetas: list[list[float]]
    sample_ids: list[int]
    weights: list[float]
    correct_metric: str = None
    bench: str = field(default=None)
    mode: str = field(default=None)


class CrossLingualMathProblem(MergingProblem):
    """
    Class for optimizating merged models in order to transfer skills across different languages.
    """

    def __init__(
        self,
        merger,
        search_df: pd.DataFrame | None,
        test_df: pd.DataFrame | None,
        lang_id: str,
        conf_pe: ConfigPE,
        lm_eval_tasks: dict[str, str] | None = None,
        n_var: int = 11,
        n_obj: int = 2,
        n_eq_constr: int = 0,
        n_ieq_constr: int = 0,
        xl=0,
        xu=1,
        eval_batch_size: int = 5,
        detect_lang=True,
        additional_templates_folder: str | None = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        merger : Merger
            The merger to use for merging models.
        search_df : pd.DataFrame
            The dataframe to use for searching for the best model.
        test_df : pd.DataFrame
            The dataframe to use for testing the best model.
        lm_eval_tasks : dict[str, str]
            Dictionary of tasks to evaluate the model on, each corresponding to a search or test phase.
        lang_id : str
            Identifier for the language being processed.
        n_var : int
            Number of variables.
        n_obj : int
            Number of objectives.
        n_eq_constr : int
            Number of equality constraints.
        n_ieq_constr : int
            Number of inequality constraints.
        xl : int
            Lower bound for the variables.
        xu : int
            Upper bound for the variables.
        eval_batch_size : int
            Batch size for evaluation.
        """

        # init the parent with the right params
        super().__init__(
            merger=merger,
            search_df=search_df,
            test_df=test_df,
            n_var=n_var,
            n_obj=n_obj,
            n_eq_constr=n_eq_constr,
            n_ieq_constr=n_ieq_constr,
            xl=xl,
            xu=xu,
            eval_batch_size=eval_batch_size,
            use_lm_eval=lm_eval_tasks is not None,
            **kwargs,
        )

        self.test_df = test_df
        self.lm_eval_tasks = lm_eval_tasks
        self.lang_id = lang_id
        self.conf_pe = conf_pe
        self.detect_lang = detect_lang
        self.additional_templates_folder = additional_templates_folder

    def metrics_4_genotype(
        self, model, tokenizer: Optional[object] = None
    ) -> Union[list[float], str]:
        """
        Method to evaluate the performance of the merged model/genotype.
        A prediction is correct if in italian and the last number in the string is the correct answer.

        Parameters
        ----------
        model : Model
            The model to evaluate.
        tokenizer : Tokenizer
            The tokenizer to use for the model. Not required if self.use_lm_eval is True.
        Returns
        -------
        list[float]
            A list of metrics that will be used to evaluate the performance of the merged model in out["F"]
        """

        # get the predictions
        device = self.device if self.device is not None else "cpu"

        df = self.test_df if self.test_mode else self.search_df

        if not self.use_lm_eval:
            df["predictions"] = get_batched_model_predictions(
                model,
                tokenizer,
                df,
                batch_size=self.eval_batch_size,
                max_token=256,
                device=device,
            )
            # remove prompt text in column 'prompt' from 'predictions'
            df["predictions"] = df.apply(
                lambda row: row["predictions"].replace(row["prompt"], ""), axis=1
            )

            # get the loss with the evaluator
            evaluator = FGMathEvaluator(language_id=self.lang_id)
            correctness = evaluator.get_correctness(df)
        else:
            evaluator = LmHarnessEvaluator(
                task_name=self.lm_eval_tasks[
                    "search" if not self.test_mode else "test"
                ][self.lang_id],
                sample_ids=self.conf_pe.sample_ids,
                correctness_metric=self.conf_pe.correct_metric,
                lang_id=self.lang_id if self.detect_lang else None,
                is_test=self.test_mode,
                additional_templates_folder=self.additional_templates_folder,
                batch_size=self.eval_batch_size,
            )
            correctness = evaluator.evaluate(model)

        est_params = PerformanceEstimationParameters(
            thetas=self.conf_pe.thetas,
            sample_weights=self.conf_pe.weights,
            sample_ids=self.conf_pe.sample_ids,
            mode=self.conf_pe.mode if not self.test_mode else "mean",
            bench=self.conf_pe.bench,
        )

        f = -(PerformanceEstimator(est_params).estimate_accuracy(correctness))

        return [f], f"Fitness value: {-f}"

    def test(
        self, genotype, base_model: Optional[Path] = None
    ) -> Union[list[float], str]:
        """
        Evaluate a model's performance on a test dataset.

        Parameters
        ----------
        genotype : list
            The genotype to evaluate.
        base_model : Path, optional
            The path to the base model to use for evaluation.

        Returns
        -------
        Union[list[float], str]
            The metrics for the model.
        """
        self.test_mode = True

        if base_model:
            if self.use_lm_eval:
                model = self.load_model(base_model)
            else:
                model, tokenizer = self.load_model(base_model)
        else:
            assert (
                len(genotype) == self.n_var
            ), f"Genotype length mismatch: expected {self.n_var}, got {len(genotype)}."
            path_to_model = self._from_array_to_genotype(genotype)

            if self.use_lm_eval:
                model = self.load_model(path_to_model)
            else:
                model, tokenizer = self.load_model(path_to_model)

        if self.use_lm_eval:
            return self.metrics_4_genotype(model)
        else:
            return self.metrics_4_genotype(model, tokenizer)


# =====================
#  MULTILINGUAL PROBLEM
# =====================
@dataclass
class ConfigMultiLingualPE:
    sample_ids: dict[str, list[int]]
    weights: dict[str, list[float]]
    thetas: list[list[float]]
    bench: str
    mode: str
    correct_metric: str = None


class MultilingualMergingProblem(MultiObjectiveMergingProblem):
    """
    Class for optimizating merged models on multiple languages.
    """

    def __init__(
        self,
        merger,
        search_df_dict: dict[str, pd.DataFrame] | None,
        test_df_dict: dict[str, pd.DataFrame] | None,
        config_pe: ConfigMultiLingualPE,
        lm_eval_tasks: dict[str, dict[str]] | None = None,
        n_var: int = 11,
        n_obj: int = 2,
        n_eq_constr: int = 0,
        n_ieq_constr: int = 0,
        xl=0,
        xu=1,
        eval_batch_size: int = 5,
        detect_lang=True,
        additional_templates_folder: str | None = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        merger : Merger
            The merger to use for merging models.
        dataframe_dictionary : dict[str, pd.DataFrame]
            Dictionary of dataframes, each corresponding to a language.
        test_df_dict : dict[str, pd.DataFrame]
            Dictionary of test dataframes, each corresponding to a language.
        lm_eval_tasks : dict[str, dict[str]] | None
            Dictionary of tasks to evaluate the model on, each corresponding to search and test.
        config_pe : ConfigMultiLingualPE
            Configuration for performance estimation.
        n_var : int
            Number of variables.
        n_obj : int
            Number of objectives.
        n_eq_constr : int
            Number of equality constraints.
        n_ieq_constr : int
            Number of inequality constraints.
        xl : int
            Lower bound for the variables.
        xu : int
            Upper bound for the variables.

        Returns
        -------
        None
        """

        # init the parent with the right params
        super().__init__(
            merger=merger,
            search_dataframes=search_df_dict,
            test_dataframes=test_df_dict,
            n_var=n_var,
            n_obj=n_obj,
            n_eq_constr=n_eq_constr,
            n_ieq_constr=n_ieq_constr,
            xl=xl,
            xu=xu,
            eval_batch_size=eval_batch_size,
            use_lm_eval=lm_eval_tasks is not None,
            **kwargs,
        )

        self.search_df_dict = search_df_dict
        self.test_df_dict = test_df_dict
        self.lm_eval_tasks = lm_eval_tasks
        self.conf_pe = config_pe
        self.detect_lang = detect_lang
        self.additional_templates_folder = additional_templates_folder

    def metrics_4_genotype(self, model, tokenizer=None) -> Union[list[float], str]:
        """
        Method to evaluate the performance of the merged model/genotype.
        A prediction is correct if in italian and the last number in the string is the correct answer.

        Parameters
        ----------
        model : Model
            The model to evaluate.
        tokenizer : Tokenizer
            The tokenizer to use for the model.
        Returns
        -------
        Union[list[float], str]
            A list of metrics that will be used to evaluate the performance of the merged model in out["F"]
        """
        device = self.device if self.device is not None else "cpu"
        dfs = self.test_df_dict if self.test_mode else self.search_df_dict

        if not self.use_lm_eval:
            # get predictions
            for k, data in dfs.items():
                data["predictions"] = get_batched_model_predictions(
                    model,
                    tokenizer,
                    data,
                    batch_size=self.eval_batch_size,
                    device=device,
                    print_output=True,
                )
                # remove prompt text in column 'prompt' from 'predictions'
                data["predictions"] = data.apply(
                    lambda row: row["predictions"].replace(row["prompt"], ""), axis=1
                )

            evaluator = MultilingualMCEvaluator(
                language_ids=list(dfs.keys()), validate_lang=True
            )
            correctness_dict = evaluator.get_correctness(dfs)

        else:
            correctness_dict = {}
            tasks = self.lm_eval_tasks["search" if not self.test_mode else "test"]
            for lang in self.conf_pe.sample_ids.keys():
                evaluator = LmHarnessEvaluator(
                    task_name=tasks[lang],
                    sample_ids=self.conf_pe.sample_ids[lang],
                    correctness_metric=self.conf_pe.correct_metric,
                    lang_id=lang if self.detect_lang else None,
                    is_test=self.test_mode,
                    additional_templates_folder=self.additional_templates_folder,
                    batch_size=self.eval_batch_size,
                )
                correctness_dict[lang] = evaluator.evaluate(model)

        # get metrics
        acc_dict = {}
        for k, correctness in correctness_dict.items():
            est_params = PerformanceEstimationParameters(
                thetas=self.conf_pe.thetas,
                sample_weights=self.conf_pe.weights[k],
                sample_ids=self.conf_pe.sample_ids[k],
                mode=self.conf_pe.mode if not self.test_mode else "mean",
                bench=self.conf_pe.bench,
            )
            perf_estimator = PerformanceEstimator(est_params)

            logger.info(f"Correctness for {k}: {correctness}")
            acc_dict[k] = perf_estimator.estimate_accuracy(correctness)

        f = [-1 * acc for acc in acc_dict.values()]
        description = "Fitness values: " + str(acc_dict)

        if not f:
            raise ValueError("No metrics were computed.")

        return f, description

    def test(
        self, genotype, base_model: Optional[Path] = None
    ) -> Union[list[float], str]:
        """
        Evaluate a model's performance on a test dataset.

        Parameters
        ----------
        genotype : list
            The genotype to evaluate.
        base_model : Path, optional
            The path to the base model to use for evaluation.

        Returns
        -------
        Union[list[float], str]
            The metrics for the model.
        """
        self.test_mode = True

        if base_model:
            if self.use_lm_eval:
                model = self.load_model(base_model)
            else:
                model, tokenizer = self.load_model(base_model)
        else:
            assert (
                len(genotype) == self.n_var
            ), f"Genotype length mismatch: expected {self.n_var}, got {len(genotype)}."
            path_to_model = self._from_array_to_genotype(genotype)

            if self.use_lm_eval:
                model = self.load_model(path_to_model)
            else:
                model, tokenizer = self.load_model(path_to_model)

        if self.use_lm_eval:
            return self.metrics_4_genotype(model)
        else:
            return self.metrics_4_genotype(model, tokenizer)


# ===============================
#  LM-EVAL MULTIOBJECTIVE PROBLEM
# ===============================
@dataclass
class ConfigLmEvalMultiObjectivePE:
    tasks: list[str]
    correct_metric: str
    sample_ids: dict[str, list[int]]
    additional_templates_folder: str | None = None

    def __post_init__(self):
        if not isinstance(self.tasks, list):
            raise ValueError("Tasks should be a list of task names.")
        if not isinstance(self.sample_ids, dict):
            raise ValueError("Sample IDs should be a dictionary of lists of integers.")
        if not isinstance(self.correct_metric, str):
            raise ValueError("Correctness metric should be a string.")
        if not isinstance(self.additional_templates_folder, (str, type(None))):
            raise ValueError("Additional templates folder should be a string or None.")


class LmEvalMultiObjectiveProblem(MultiObjectiveMergingProblem):
    """
    Class for evolving merged models.
    """

    def __init__(
        self,
        config: ConfigLmEvalMultiObjectivePE,
        merger,
        n_var: int = 11,
        n_obj: int = 2,
        n_eq_constr: int = 0,
        n_ieq_constr: int = 0,
        xl=0,
        xu=1,
        **kwargs,
    ):
        super().__init__(
            merger=merger,
            n_var=n_var,
            n_obj=n_obj,
            n_eq_constr=n_eq_constr,
            n_ieq_constr=n_ieq_constr,
            xl=xl,
            xu=xu,
            search_dataframes=None,
            test_dataframes=None,
            use_lm_eval=True,
            **kwargs,
        )

        self.config = config

    def metrics_4_genotype(self, model) -> Union[list[float], str]:
        """
        Method to evaluate the performance of the merged model/genotype.

        Parameters
        ----------
        model : Model
            The model to evaluate.
        Returns
        -------
        Union[list[float], str]
            A list of metrics that will be used to evaluate the performance of the merged model in out["F"]
        """

        correctness_dict = {}
        for task in self.config.tasks:
            evaluator = LmHarnessEvaluator(
                task_name=task,
                sample_ids=self.config.sample_ids[task],
                correctness_metric=self.config.correct_metric,
                is_test=self.test_mode,
                additional_templates_folder=self.config.additional_templates_folder,
                batch_size=self.eval_batch_size,
            )
            correctness_dict[task] = evaluator.evaluate(model)

        # get metrics
        acc_dict = {}
        for k, correctness in correctness_dict.items():
            est_params = PerformanceEstimationParameters(
                thetas=None,
                sample_weights=np.ones(len(self.config.sample_ids[k]))
                / len(self.config.sample_ids[k]),
                sample_ids=self.config.sample_ids[k],
                mode="mean",
                bench=None,
            )
            perf_estimator = PerformanceEstimator(est_params)

            logger.info(f"Correctness for {k}: {correctness}")
            acc_dict[k] = perf_estimator.estimate_accuracy(correctness)

        f = [-1 * acc for acc in acc_dict.values()]
        description = "Fitness values: " + str(acc_dict)

        if not f:
            raise ValueError("No metrics were computed.")

        return f, description

    def test(
        self, genotype, base_model: Optional[Path] = None
    ) -> Union[list[float], str]:
        """
        Evaluate a model's performance on a test dataset.
        Parameters
        ----------
        genotype : list
            The genotype to evaluate.
        base_model : Path, optional
            The path to the base model to use for evaluation.
        Returns
        -------
        Union[list[float], str]
            The metrics for the model.
        """
        self.test_mode = True
        if base_model:
            model = self.load_model(base_model)
        else:
            assert (
                len(genotype) == self.n_var
            ), f"Genotype length mismatch: expected {self.n_var}, got {len(genotype)}."
            path_to_model = self._from_array_to_genotype(genotype)

        model = self.load_model(path_to_model)

        return self.metrics_4_genotype(model)



import logging
import os
from pathlib import Path

import git

logger = logging.getLogger(__name__)

try:
    PROJECT_ROOT = Path(
        git.Repo(Path.cwd(), search_parent_directories=True).working_dir
    )
except git.exc.InvalidGitRepositoryError:
    PROJECT_ROOT = Path.cwd()

try:
    CACHE_DIR = (
        Path(git.Repo(Path.cwd(), search_parent_directories=True).working_dir) / "cache"
    )
except git.exc.InvalidGitRepositoryError as e:
    # should use the predefined hugging-face directory
    logger.error(e)
    CACHE_DIR = None
    logger.debug("\n setting default hugging-face cache directory")

if "mergenetic" not in str(PROJECT_ROOT):

    logger.warning(f"PROJECT_ROOT IS HARCODED AT {PROJECT_ROOT}")
    CACHE_DIR = PROJECT_ROOT / "cache"

os.environ["PROJECT_ROOT"] = str(PROJECT_ROOT)
os.environ["CACHE_DIR"] = str(CACHE_DIR)

#############################################
# importing the necessary modules to set up the project
#############################################
from .utils import *



from logging import getLogger
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pymoo.core.algorithm import Algorithm
from pymoo.optimize import minimize

from mergenetic.optimization import MergingProblem

logger = getLogger(__name__)


class Searcher:
    def __init__(
        self,
        problem: MergingProblem,
        algorithm: Algorithm,
        results_path: str,
        n_iter: int,
        seed: int,
        run_id: str,
        verbose: bool = True,
    ):
        """
        Initializes the Seaercher with a problem, an algorithm, and a path to save results.

        :param problem: An instance of a Problem subclass.
        :param algorithm: An instance of an Algorithm class.
        :param results_path: Path to save the optimization results.
        :param n_iter: Number of iterations to run the optimization algorithm.
        :param seed: Random seed for reproducibility.
        :param verbose: Whether to print optimization progress
        """
        self.problem = problem
        self.algorithm = algorithm
        self.results_path = Path(results_path)
        self.n_iter = n_iter
        self.seed = seed
        self.verbose = verbose
        self.run_id = run_id

    def search(self) -> pd.DataFrame:
        """Executes the search algorithm and saves the results to a CSV file."""
        result = minimize(
            self.problem,
            self.algorithm,
            ("n_iter", self.n_iter),
            seed=self.seed,
            verbose=self.verbose,
        )

        self.result_X = result.X / 10 if self.problem.discrete else result.X
        self.result_F = result.F
        logger.info(f"Best solution found: {result.X}. Best function value: {result.F}")

        if hasattr(self.problem, "results_df"):
            if isinstance(self.problem.results_df, pd.DataFrame):
                search_path = self.results_path / Path(f"{self.run_id}.csv")
                self.problem.results_df.to_csv(search_path)
                return self.problem.results_df
            elif isinstance(self.problem.results_df, dict):
                for key, value in self.problem.results_df.items():
                    search_path = self.results_path / Path(f"{self.run_id}_{key}.csv")
                    value.to_csv(search_path)
                return self.problem.results_df
            else:
                logger.error("Problem 'results_df' is not a DataFrame or a dictionary.")
                return None
        else:
            logger.info("Problem does not have 'results_df' to save.")
            return None

    def test(self):
        # check that a test dataframe was given
        # check that the path is valid
        logger.info("Starting the test...")

        if len(self.result_X.shape) > 1:
            for x in self.result_X:
                # check that is a numpy array
                assert type(x) == np.ndarray, "not a numpy array"
                fit, _ = self.problem.test(x)
                results_df = self.problem.get_data()
                if isinstance(self.problem.results_df, pd.DataFrame):
                    path = self.results_path / Path(f"{self.run_id}_test.csv")
                    logger.info(f"The genotype {x} got fitness {fit}")
                    results_df.to_csv(
                        self.results_path / Path(f"{self.run_id}_test.csv")
                    )
                else:
                    for key, value in results_df.items():
                        path = self.results_path / Path(f"{self.run_id}_test_{key}.csv")
                        logger.info(f"The genotype {x} got fitness {fit} on {key}.")
                        value.to_csv(path)
            logger.info("TEST MODE COMPLETED")
        else:
            # check that is a numpy array
            assert type(self.result_X) == np.ndarray, "not a numpy array"
            accuracy, _ = self.problem.test(self.result_X)
            logger.info(f"The genotype {self.result_X} got accuracy {accuracy}")
            results_df = self.problem.get_data()

            if isinstance(self.problem.results_df, pd.DataFrame):
                path = self.results_path / Path(f"{self.run_id}_test.csv")
                results_df.to_csv(path)
            else:
                for key, value in results_df.items():
                    path = self.results_path / Path(f"{self.run_id}_{key}_test.csv")
                    value.to_csv(path)
            logger.info("TEST MODE COMPLETED")

    def visualize_results(self) -> None:
        """Plots metrics and phenotypes over the optimization steps from the results_df."""
        if not hasattr(self.problem, "results_df"):
            raise AttributeError(
                "Problem does not have 'results_df' to visualize results."
            )

        df = self.problem.results_df
        metrics = [col for col in df.columns if "objective" in col]
        phenotypes = [col for col in df.columns if "phenotype" in col]

        for metric in metrics:
            plt.figure(figsize=(10, 4))
            plt.plot(df["step"], df[metric], marker="o", linestyle="-")
            plt.title(f"Metric: {metric}")
            plt.xlabel("Step")
            plt.ylabel(metric)
            plt.grid(True)
            plt.show()

        for phenotype in phenotypes:
            plt.figure(figsize=(10, 4))
            plt.plot(df["step"], df[phenotype], marker="x", linestyle="--")
            plt.title(f"Phenotype: {phenotype}")
            plt.xlabel("Step")
            plt.ylabel(phenotype)
            plt.grid(True)
            plt.show()



from .merging_problem import MergingProblem, MultiObjectiveMergingProblem
from .predefined_problems import (
    CrossLingualMathProblem,
    LmEvalMultiObjectiveProblem,
    MultilingualMergingProblem,
)



from logging import getLogger

import numpy as np
import pandas as pd
from lm_eval.api.task import ConfigurableTask
from lm_eval.evaluator import simple_evaluate
from lm_eval.models.vllm_causallms import VLLM
from lm_eval.tasks import TaskManager

from .evaluator import LanguageDetector

logger = getLogger(__name__)


class LmHarnessEvaluator:
    """
    Evaluates the performance of a language model on a given task.

    Parameters
    ----------
    task_name : str
        Name of the task to evaluate.
    """

    task: ConfigurableTask
    task_manager: TaskManager = None
    sample_ids: list[int] | None
    lang_detector: LanguageDetector | None
    lang_id: str | None

    def __init__(
        self,
        task_name: str,
        correctness_metric: str = "exact_match",
        sample_ids: list[int] | None = None,
        lang_id: str | None = None,
        is_test: bool = False,
        additional_templates_folder: str | None = None,
        batch_size: int = 32,
    ) -> None:

        self.task_manager = TaskManager(include_path=additional_templates_folder)
        self.task: ConfigurableTask = self.get_task(task_name)

        self.sample_ids = sample_ids

        if sample_ids is not None and len(sample_ids) > 0:
            if is_test:
                # take all excluding sample_ids
                ids = np.arange(len(self.task.dataset["test"]))
                self.task.dataset["test"] = self.task.dataset["test"].select(
                    np.setdiff1d(ids, sample_ids)
                )

            else:
                self.task.dataset["test"] = self.task.dataset["test"].select(sample_ids)
                logger.info("Sample ids provided. Using the specified sample ids.")
                logger.info(f"Selected samples: {self.task.dataset['test']}")
        else:
            logger.info("No sample ids provided. Using the entire dataset.")

        if self.task.OUTPUT_TYPE == "multiple_choice":
            # disable language detection for multiple choice tasks
            logger.warning("Disabling language detection for multiple choice tasks.")
            lang_id = None

        try:
            self.lang_detector = (
                LanguageDetector([lang_id]) if lang_id is not None else None
            )
        except Exception as e:
            logger.warning(f"Language detection is disabled. Error: {e}")
            self.lang_detector = None

        self.correctness_metric = correctness_metric
        self.task_nm = task_name
        self.lang_id = lang_id
        self.batch_size = batch_size

    def get_task(self, task_name: str) -> ConfigurableTask:
        """
        Returns a task for the specified task name.

        Parameters
        ----------
        task_name : str
            Name of the task.

        Returns
        -------
        ConfigurableTask
            Task for the specified task.
        """
        return self.task_manager.load_task_or_group(task_name)[task_name]

    def evaluate(self, model: VLLM) -> list[float]:
        """
        Evaluates the performance of the model on the specified task.

        Parameters
        ----------
        model : VLLM
            The model to evaluate.

        Returns
        -------
        dict
            Evaluation results.
        """
        results = simple_evaluate(model, tasks=[self.task], batch_size=self.batch_size)

        def get_responses(results):
            answers = []
            logger.info(f"Num of Answers: {len(results['samples'][self.task_nm])}")
            for sample in results["samples"][self.task_nm]:
                if isinstance(sample["resps"], list) and len(sample["resps"]) > 0:
                    if (
                        isinstance(sample["resps"][0], list)
                        and len(sample["resps"][0]) > 0
                    ):
                        # flatten to a single list
                        list_resp = [
                            item for sublist in sample["resps"] for item in sublist
                        ]
                        answers.append(
                            {
                                "id": sample["doc_id"],
                                "correctness": sample[self.correctness_metric],
                                "model_answers": list_resp,
                            }
                        )
                    else:
                        answers.append(
                            {
                                "id": sample["doc_id"],
                                "correctness": sample[self.correctness_metric],
                                "model_answers": sample["resps"][0],
                            }
                        )
                else:
                    answers.append(
                        {
                            "id": sample["doc_id"],
                            "correctness": sample[self.correctness_metric],
                            "model_answers": sample["resps"],
                        }
                    )
            return answers

        self.data = pd.DataFrame(get_responses(results))

        # let's filter the answers by sample_ids
        self.data = (
            self.data[self.data["id"].isin(self.sample_ids)]
            if self.sample_ids is not None
            else self.data
        )

        # if the number of answers is more than len(sample_ids), then we need to filter the answers
        # by randomly picking one with the same id

        if self.sample_ids is not None and len(self.data) > len(self.sample_ids):
            self.data = self.data.groupby("id").sample(n=1).reset_index(drop=True)

        logger.info(f"Samples from model answers: {self.data['model_answers'].head(5)}")

        if self.lang_detector is None:
            logger.info(
                f"Language detection is disabled. Fitness: {self.data['correctness'].mean()}"
            )
            return self.data["correctness"]
        else:
            if isinstance(self.data["model_answers"][0], str):
                self.data["language"] = self.data["model_answers"].apply(
                    lambda x: self.lang_detector._get_language(
                        x[0] if not isinstance(x, str) else x
                    )
                )
                self.data["is_language_correct"] = self.data["language"] == (
                    "__label__" + self.lang_id
                )
            else:
                self.data["language"] = self.data["model_answers"].apply(
                    lambda x: [
                        self.lang_detector._get_language(
                            y[0] if not isinstance(y, str) else y
                        )
                        for y in x
                    ]
                )
                self.data["is_language_correct"] = self.data["language"].apply(
                    lambda x: avg_lang_correctness(self.lang_id, x)
                )

            logger.info(
                f"Language detection is enabled. Fitness: {(self.data['is_language_correct'] * self.data['correctness']).mean()}"
            )
            return self.data["is_language_correct"] * self.data["correctness"]

    def get_data(self) -> pd.DataFrame:
        """
        Returns the evaluation data.

        Returns
        -------
        pd.DataFrame
            Evaluation data.
        """
        return self.data


def avg_lang_correctness(lang_id: str, langs: list[str]) -> float:
    """
    Returns the average correctness of the language detection.

    Parameters
    ----------
    langs : list[str]
        List of languages.

    Returns
    -------
    float
        Average correctness of the language detection. Returns np.nan if langs is empty.
    """
    if not langs:  # Check if the list is empty
        return (
            np.nan
        )  # Return NaN directly to avoid np.mean warning, aligns with test expectation
    return np.mean([lang == "UNK" or lang == f"__label__{lang_id}" for lang in langs])



from dataclasses import dataclass
from typing import List

import numpy as np
import pandas as pd

from mergenetic.estimator.utils import estimate_fitness

# ==========================
#  DATA CLASSES FOR ESTIMATION
# ==========================


@dataclass
class PerformanceEstimationParameters:
    """Stores parameters for accuracy estimation."""

    thetas: List[np.array]
    sample_weights: np.array
    sample_ids: np.array
    bench: str
    mode: str  # "mean", "weighted", "mpirt", "gmpirt"


# ==========================
#  PERFORMANCE ESTIMATOR
# ==========================


class PerformanceEstimator:
    """
    Evaluates the accuracy of a model using anchored evaluation techniques.
    """

    est_parameters: PerformanceEstimationParameters

    def __init__(self, est_parameters: PerformanceEstimationParameters) -> None:
        """
        Initializes the performance estimator.

        Parameters
        ----------
        est_parameters : PerformanceEstimationParameters
            Parameters for accuracy estimation.
        """
        self.est_parameters = est_parameters

    def __str__(self) -> str:
        return "Performance Estimator for LLM predictions against mathematical word problems."

    def estimate_accuracy(self, correctness: pd.Series) -> float:
        """
        Estimates the accuracy of the model based on the correctness of the predictions.

        Parameters
        ----------
        correctness : pd.Series
            Series containing the correctness of the predictions.

        Returns
        -------
        float
            Estimated accuracy of the model.
        """
        # Convert boolean results to integers (1 = correct, 0 = incorrect)
        y = correctness.astype(int).values

        match self.est_parameters.mode:
            case "mean":
                return y.mean()
            case "mpirt" | "gmpirt":
                if (
                    self.est_parameters.mode == "gmpirt"
                    and len(self.est_parameters.thetas) == 1
                ):
                    raise ValueError("GMPIRT requires multiple thetas.")

                # Estimate accuracy using anchoring methods
                estimates = estimate_fitness(
                    y,
                    self.est_parameters.thetas,
                    self.est_parameters.bench,
                    self.est_parameters.sample_ids,
                    self.est_parameters.sample_weights,
                )

                # Adjust key based on number of thetas for pirt/gpirt vs mpirt/gmpirt
                mode_key = self.est_parameters.mode
                if len(self.est_parameters.thetas) == 1:
                    if mode_key == "mpirt":
                        mode_key = "pirt"
                    elif (
                        mode_key == "gmpirt"
                    ):  # This case is already guarded by the ValueError above
                        mode_key = "gpirt"

                return estimates[mode_key]
            case "weighted":
                return (self.est_parameters.sample_weights * y).sum()

        raise ValueError(
            f"Invalid mode '{self.est_parameters.mode}' for accuracy estimation."
        )



import re
from logging import getLogger
from typing import List, Optional, Union

import pandas as pd

from .evaluator import BaseEvaluator, LanguageDetector

logger = getLogger(__name__)


def extract_numbers(
    sentence: str, only_last_number: bool = True
) -> Optional[Union[str, List[str]]]:
    """
    Extracts numbers from a sentence.

    Parameters
    ----------
    sentence : str
        Input sentence.
    only_last_number : bool
        If True, return only the last number found.

    Returns
    -------
    str | list | None
        Extracted number(s) or None if no numbers found.
    """
    # Use a robust regex to find all sequences of digits, possibly with a decimal point,
    # ensuring they are whole numbers/decimals (bounded by word boundaries).
    all_numbers = re.findall(r"\b\d+(?:\.\d+)?\b", sentence)

    if not all_numbers:
        return None

    if only_last_number:
        return all_numbers[-1]
    else:
        return all_numbers


# ==================================
#  MULTIPLE CHOICE EVALUATOR
# ==================================


class MCEvaluator(BaseEvaluator):
    """
    Evaluates the capabilities of an LLM for Multiple Choice questions in a specified language.
    """

    lang_detector: Optional[LanguageDetector]

    def __init__(self, language_id: str | None = "it") -> None:
        """
        Parameters
        ----------
        language_id : str
            Target language ID (default: "it" for Italian).
        """

        self.language_id = language_id
        self.lang_detector = None

        if self.language_id:
            try:
                self.lang_detector = LanguageDetector([self.language_id])
            except Exception as e:
                logger.warning(f"Language detection disabled: {e}")
                self.lang_detector = None

    def get_correctness(self, dataframe: pd.DataFrame) -> pd.Series:
        """
        Computes the correctness of model answers for Multiple Choice questions.

        Parameters
        ----------
        dataframe : pd.DataFrame
            DataFrame containing the model predictions and the correct answers.

        Returns
        -------
        pd.Series
            Returns the correctness of the answers.
        """
        self._validate_dataframe(dataframe)
        self.data = dataframe.copy()

        if not self.lang_detector:  # Check if lang_detector is available
            self.data = self.data.assign(
                filtered_answer=self.data["predictions"].apply(extract_numbers),
            )
            self.data["correctness"] = self.data["filtered_answer"] == self.data[
                "answer"
            ].astype(str)
            return self.data["correctness"]
        else:
            # Use list comprehension for language detection to avoid potential apply issues with mocks
            langs = [
                self.lang_detector._get_language(p) for p in self.data["predictions"]
            ]
            self.data = self.data.assign(
                language=pd.Series(langs, index=self.data.index),
                filtered_answer=self.data["predictions"].apply(extract_numbers),
            )
            self.data = self.data.assign(
                is_language_correct=lambda df: (
                    df["language"] == f"__label__{self.language_id}"
                )
                | (df["language"] == "UNK"),
                is_answer_correct=lambda df: df["filtered_answer"]
                == df["answer"].astype(str),
            )
            self.data["correctness"] = (
                self.data["is_language_correct"] & self.data["is_answer_correct"]
            )
            return self.data["correctness"]

    def get_data(self) -> pd.DataFrame:
        """
        Returns the DataFrame with the results.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the results of the evaluation.
        """
        return self.data


# ==================================
#  MATH FREE GENERATION EVALUATOR
# ==================================


class FGMathEvaluator(BaseEvaluator):
    """
    Evaluates the capabilities of an LLM in solving math word problems in a specified language.
    """

    lang_detector: Optional[LanguageDetector]  # Allow Optional for consistency

    def __init__(self, language_id: str | None = None) -> None:
        """
        Parameters
            ----------
        language_id : str
            Target language ID (default: "it" for Italian).
        """

        self.language_id = language_id
        self.lang_detector = None  # Initialize to None

        if self.language_id:  # Use self.language_id here
            try:
                self.lang_detector = LanguageDetector([self.language_id])
            except Exception as e:
                logger.warning(f"Language detection for FGMathEvaluator disabled: {e}")
                self.lang_detector = None

    @staticmethod
    def _extract_answer(text: str) -> str:
        """
        Extracts the last number from the model's prediction.

        Parameters
        ----------
        text : str
            The text from which to extract the answer.

        Returns
        -------
        str
            The extracted number, or 'None' if no match is found.
        """
        matches = re.findall(r"\d+", text)
        return matches[-1] if matches else "None"

    def get_correctness(
        self,
        dataframe: pd.DataFrame,
    ) -> pd.Series:
        """
        Computes the accuracy of the model for math word problems.

        Parameters
        ----------
        return_dataframe : bool, optional
            If True, return the full evaluation DataFrame.

        Returns
        -------
        dict[str, float] | tuple[dict[str, float], dict[str, pd.DataFrame]]
            Accuracy for each language, and optionally the evaluation DataFrame.
        """
        self._validate_dataframe(dataframe)

        self.data = dataframe.copy()
        self.data["predictions_filtered"] = self.data["predictions"].apply(
            self._extract_answer
        )
        self.data["correctness"] = self.data["predictions_filtered"] == self.data[
            "answer"
        ].astype(str)

        if self.lang_detector:  # Check if lang_detector is available
            # Use list comprehension for language detection
            langs = [
                self.lang_detector._get_language(p) for p in self.data["predictions"]
            ]
            self.data["language"] = pd.Series(langs, index=self.data.index)

            self.data["evaluation_log"] = self.data.apply(
                lambda row: f"Predicted: {row.predictions_filtered}, Actual: {row.answer}, Correct: {row.correctness}\nFull Answer: {row.predictions}, Language: {row.language}",
                axis=1,
            )

            # Combine correctness with language correctness
            lang_correct = (self.data["language"] == f"__label__{self.language_id}") | (
                self.data["language"] == "UNK"
            )
            return self.data["correctness"] & lang_correct
        else:
            self.data["evaluation_log"] = self.data.apply(
                lambda row: f"Predicted: {row.predictions_filtered}, Actual: {row.answer}, Correct: {row.correctness}\nFull Answer: {row.predictions}",
                axis=1,
            )
            return self.data["correctness"]

    def get_data(self) -> pd.DataFrame:
        """
        Returns the DataFrame with the results.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the extracted numbers.
        """
        return self.data



import logging
import shutil
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

import torch  # Ensure torch is imported if used for cuda.is_available
import yaml  # Ensure yaml is imported if used for safe_load
from mergekit.config import MergeConfiguration
from mergekit.merge import MergeOptions, run_merge

from mergenetic import clean_gpu

logger = logging.getLogger(__name__)


class Merger(ABC):
    def __init__(
        self,
        run_id: str,
        path_to_store_yaml: str,
        path_to_store_merged_model: str,
        dtype: str,
        **kwargs: Any,
    ) -> None:
        """
        Abstract class for merging models. It is used to create a configuration file for merging models according to a given technique (DARE, TIES, etc.).
        It requires all the information for merging through mergekit library according to the given merging technique. Two type of information are used:
            1. Static information common to all merged models with that technique (e.g., base model, dtype, etc.); these are passed through the class init.
            2. Dynamic information specific to each merging operation (e.g., weights, densities, etc.); these are passed through the method for merging.

        Parameters
        ----------
        run_id : str
            is the id of the run that will be used to store all the yaml configuration files for merging.
        path_to_store_yaml : str
            is the path to the folder where the yaml configuration files for merging will be stored.
        path_to_store_merged_model : str
            is the path to the folder where the merged models will be stored.
        dtype : str
            is the data type of the models (e.g., float16, float32, etc.).
        **kwargs :
            is the list of keyword arguments that will be used to create the configuration file.

        Returns
        -------
        None
        """

        # init ABC (parent class)
        super().__init__()

        # storing all the paths
        self.run_id = run_id
        # Ensure path_to_store_yaml is the full file path
        self.path_to_store_yaml = Path(path_to_store_yaml) / "config.yaml"
        self.path_to_store_merged_model = Path(path_to_store_merged_model) / self.run_id

        # storing the data type
        self.dtype = dtype
        # Handle other kwargs if necessary, e.g., for specific merger types

    @abstractmethod
    def create_individual_configuration(self, *args: Any, **kwargs: Any) -> Path:
        """
        Creates the individual configuration file for the merger.
        """
        pass

    def check_and_delete_yaml(self) -> None:
        """
        Checks if the YAML configuration file exists and deletes it.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        if self.path_to_store_yaml.exists():  # check if the file exists
            self.path_to_store_yaml.unlink()  # Deletes the file
            logger.info(f"Deleted: {self.path_to_store_yaml}")
        else:
            logger.info(f"No file found at: {self.path_to_store_yaml}")

    def _delete_merged_model_local(self) -> None:
        """
        Deletes the merged model from the local storage.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """

        # Check if the path exists before attempting to remove it
        if self.path_to_store_merged_model.exists():

            # Use shutil.rmtree to remove the directory
            shutil.rmtree(self.path_to_store_merged_model)
            logger.info(
                f"Deleted folder and all contents: {self.path_to_store_merged_model}"
            )

        else:

            logger.info(f"The folder does not exist: {self.path_to_store_merged_model}")

    def merge_model_from_configuration(self, path_to_yaml: Path) -> Path:
        """
        Merges the model from the configuration file.

        Parameters
        ----------
        path_to_yaml : Path
            is the path to the yaml configuration file created.

        Returns
        -------
        path_to_store_merged_model : : Path or str
             is the path to the merged model created.
        """
        logger.info("Starting model merging process...")
        clean_gpu()
        self._delete_merged_model_local()

        with open(path_to_yaml, "r", encoding="utf-8") as f:
            config_data = yaml.safe_load(f)

        cfg = MergeConfiguration.model_validate(config_data)

        options = MergeOptions(
            cuda=torch.cuda.is_available(),
            copy_tokenizer=True,
            lazy_unpickle=True,
            low_cpu_memory=True,
        )

        logger.info(f"Running merge with configuration: {cfg}")
        logger.info(f"Output path for merged model: {self.path_to_store_merged_model}")
        logger.info(f"Merge options: {options}")

        # Ensure the output path is passed as the second positional argument
        run_merge(cfg, str(self.path_to_store_merged_model), options=options)

        logger.info(f"Model merged successfully at {self.path_to_store_merged_model}")
        clean_gpu()
        return self.path_to_store_merged_model



import gc
import logging
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import numpy as np
import torch
from tqdm import tqdm
from transformers import AutoModelForCausalLM, AutoTokenizer, BitsAndBytesConfig

logger = logging.getLogger(__name__)


@dataclass
class Config:
    """
    A class required to run the evolve search. It allows to set the parameters for the evolution.
    """

    # Required parameters
    pop_size: int = None
    n_iter: int = None
    path_to_store_config: str = None
    path_to_store_merged_model: str = None
    datasets: Dict[str, str] = None
    run_id: str = None
    bench: str = None
    models: Dict[str, str] = None
    task_type: str = None  # FG_MATH or MULTIPLECHOICE
    langs: List[str] = None

    # Optional parameters
    seed: int = 420
    n_samples: int = 20
    mode: str = "gmpirt"
    dtype: str = "bfloat16"
    device: Optional[str] = "cuda:1"
    base_model: Optional[Any] = None
    force_evaluation: bool = False
    eval_batch_size: int = 64


@dataclass
class ConfigLmEval(Config):
    """
    A class required to run the evolve search.
    It allows to set the parameters for the evolution and evaluation with lm-eval-harness.
    """

    metric: Optional[Any] = None
    tasks: Dict[str, Dict[str, Dict[str, str]]] = None
    additional_templates_folder: Optional[str] = None
    load_in_4bit: bool = True


@dataclass
class ConfigMultiObjective:
    """
    **Only for LM-EVAL.**
    A class required to run the evolve search in a generic multi-objective setting.
    It allows to set the parameters for the evolution and evaluation with lm-eval-harness.
    """

    # Required parameters
    run_id: str = None
    n_iter: int = None
    pop_size: int = None
    metric: Optional[Any] = None
    models: List[str] = None
    path_to_store_config: str = None
    path_to_store_merged_model: str = None
    additional_templates_folder: Optional[str] = None
    tasks: List[str] = None

    # Optional parameters
    seed: int = 420
    n_samples: int = 20
    dtype: str = "bfloat16"
    eval_batch_size: int = 32
    device: Optional[str] = "cuda"
    base_model: Optional[Any] = None
    force_evaluation: bool = False
    load_in_4bit: bool = True


############################################################################################################
# UTILS 4 MODELS GENERATION AND MANAGEMENT                                                                 #
############################################################################################################


def get_batched_model_predictions(
    model,
    tokenizer,
    df,
    batch_size=32,
    randomness=False,
    device="cuda",
    max_token=1024,
    print_output=False,
    apply_chat=False,
    dataset_nm=None,
    custom_prompt_template=False,
):

    # Check if prompt column is available in the DataFrame
    assert "prompt" in df.columns, "prompt column not found in the dataframe"

    if custom_prompt_template:
        df["prompt"] = df["prompt"].apply(
            lambda x: custom_prompt_template.format(input=x)
        )

    else:
        if not apply_chat or ("arc_test" in dataset_nm):
            df["prompt"] = df["prompt"].apply(lambda x: f"Question: {x}\nAnswer:")

    # Transform data into list of questions
    list_data = df["prompt"].tolist()

    # Batch the list of questions
    assert batch_size > 0, "batch size must be greater than 0"
    batched_list_data = [
        list_data[i : i + batch_size] for i in range(0, len(list_data), batch_size)
    ]

    # Ensure the tokenizer has a pad token
    if tokenizer.pad_token is None:
        if tokenizer.eos_token:
            tokenizer.pad_token = tokenizer.eos_token
        else:
            tokenizer.add_special_tokens({"pad_token": "[PAD]"})
            model.resize_token_embeddings(len(tokenizer))

    final_answers = []

    # Disable gradient calculation to reduce memory usage
    with torch.no_grad():
        # Loop over the batches
        for batch in tqdm(batched_list_data, desc="Processing batches"):
            # Prepare model inputs
            questions = batch  # Directly use the prompts
            # Tokenize inputs
            if apply_chat:
                # Prepare model inputs
                questions = [
                    [{"role": "user", "content": prompt}] for prompt in questions
                ]
                questions = [
                    tokenizer.apply_chat_template(
                        prompt, tokenize=False, system_message=""
                    )
                    for prompt in questions
                ]

            logger.info(f"First question after processing: {questions[0]}")

            model_inputs = tokenizer(
                questions,
                return_tensors="pt",
                padding=True,
                truncation=True,
                max_length=max_token,
            ).to(device)

            # check that there are no NaNs in the model inputs
            if torch.isnan(model_inputs.input_ids).any():
                raise ValueError("NaNs in model inputs input IDs")

            if torch.isnan(model_inputs.attention_mask).any():
                raise ValueError("NaNs in model inputs attention mask")
            try:
                # Generate answers
                if custom_prompt_template:
                    generated_ids = model.generate(
                        **model_inputs,
                        max_new_tokens=max_token,
                        temperature=0,  # Control randomness
                        top_p=1.0,  # Control randomness
                        repetition_penalty=1.0,  # Control repetition
                        eos_token_id=tokenizer.eos_token_id,
                    )

                else:
                    generated_ids = model.generate(
                        **model_inputs,
                        max_new_tokens=max_token,
                        do_sample=randomness,  # Control randomness
                        eos_token_id=tokenizer.eos_token_id,
                    )

                # Decode generated IDs to get answers
                answers = tokenizer.batch_decode(
                    generated_ids, skip_special_tokens=True
                )
            except Exception as e:
                logger.error(f"Error during model generation: {e}")
                answers = ["GENERATION FAILED"] * len(questions)
            # Append answers to final list
            final_answers.extend(answers)

            # Optionally print prompts and their corresponding answers
            if print_output:
                for prompt, answer in zip(batch, answers):
                    logger.debug(f"Prompt: {prompt}\nAnswer: {answer}\n---")

            # Clean up to free memory

            model_inputs = model_inputs.to("cpu")
            try:
                generated_ids = generated_ids.to("cpu")
            except (
                AttributeError
            ):  # Handle cases where generated_ids might not have .to()
                pass
            torch.cuda.empty_cache()
    logger.debug(final_answers)
    return final_answers


# cleaner for gpu
def clean_gpu():
    torch.cuda.empty_cache()
    gc.collect()


# loading model on automatically chosen device
def loading_model_tokenizer_gpu_only_4bit_local(path):
    if not torch.cuda.is_available():
        raise RuntimeError("cuda is off :(")

    quant_config = BitsAndBytesConfig(
        load_in_4bit=True,
        bnb_4bit_quant_type="nf4",
        bnb_4bit_use_double_quant=True,
        bnb_4bit_compute_dtype=torch.bfloat16,
    )

    logger.debug(f"Loading model from: {path}")

    model = AutoModelForCausalLM.from_pretrained(
        path, device_map="cuda:0", quantization_config=quant_config
    )
    model.config.use_cache = True
    model.eval()

    tokenizer = AutoTokenizer.from_pretrained(path, padding_side="left")

    return model, tokenizer


############################################################################################################
# UTILS 4 TINYBENCHMARK: https://github.com/felipemaiapolo/tinyBenchmarks/blob/main/tutorials/utils.py     #
############################################################################################################


def prepare_data(scenarios, data):
    """
    Prepare the data by determining the positions of items within each scenario and subscenario.

    Parameters:
    - scenarios: A dictionary mapping each scenario to its subscenarios.
    - data: A dictionary containing correctness data for each subscenario.

    Returns:
    - scenarios_position: A dictionary mapping each scenario to the positions of its items.
    - subscenarios_position: A nested dictionary mapping each scenario and subscenario to the positions of its items.
    """

    i = 0
    subscenarios_position = {}

    # Iterate through each chosen scenario and its subscenarios to record item positions
    for scenario in scenarios.keys():
        subscenarios_position[scenario] = {}
        for sub in scenarios[scenario]:
            subscenarios_position[scenario][sub] = []
            for j in range(data["data"][sub]["correctness"].shape[0]):
                subscenarios_position[scenario][sub].append(i)
                i += 1

    # Prepare a simplified mapping of scenarios to their item positions
    scenarios_position = {}
    for scenario in scenarios.keys():
        scenarios_position[scenario] = []
        for key in subscenarios_position[scenario].keys():
            scenarios_position[scenario] += subscenarios_position[scenario][key]
    return scenarios_position, subscenarios_position


def create_responses(scenarios, data):
    """
    Create a matrix of responses for the chosen scenarios.

    Parameters:
    - scenarios: A dictionary mapping each scenario to its subscenarios.
    - data: A dictionary containing correctness data for each subscenario.

    Returns:
    - A numpy array of responses for the chosen scenarios.
    """

    responses = [
        np.vstack([data["data"][sub]["correctness"] for sub in scenarios[scenario]]).T
        for scenario in scenarios.keys()
    ]
    responses = np.hstack(responses)
    return responses


scenarios = {
    "harness_truthfulqa_mc_0": ["harness_truthfulqa_mc_0"],
    "gsm8k": ["harness_gsm8k_5"],
    "winogrande": ["harness_winogrande_5"],
    "arc": ["harness_arc_challenge_25"],
    "hellaswag": ["harness_hellaswag_10"],
    "mmlu": [
        "harness_hendrycksTest_abstract_algebra_5",
        "harness_hendrycksTest_anatomy_5",
        "harness_hendrycksTest_astronomy_5",
        "harness_hendrycksTest_business_ethics_5",
        "harness_hendrycksTest_clinical_knowledge_5",
        "harness_hendrycksTest_college_biology_5",
        "harness_hendrycksTest_college_chemistry_5",
        "harness_hendrycksTest_college_computer_science_5",
        "harness_hendrycksTest_college_mathematics_5",
        "harness_hendrycksTest_college_medicine_5",
        "harness_hendrycksTest_college_physics_5",
        "harness_hendrycksTest_computer_security_5",
        "harness_hendrycksTest_conceptual_physics_5",
        "harness_hendrycksTest_econometrics_5",
        "harness_hendrycksTest_electrical_engineering_5",
        "harness_hendrycksTest_elementary_mathematics_5",
        "harness_hendrycksTest_formal_logic_5",
        "harness_hendrycksTest_global_facts_5",
        "harness_hendrycksTest_high_school_biology_5",
        "harness_hendrycksTest_high_school_chemistry_5",
        "harness_hendrycksTest_high_school_computer_science_5",
        "harness_hendrycksTest_high_school_european_history_5",
        "harness_hendrycksTest_high_school_geography_5",
        "harness_hendrycksTest_high_school_government_and_politics_5",
        "harness_hendrycksTest_high_school_macroeconomics_5",
        "harness_hendrycksTest_high_school_mathematics_5",
        "harness_hendrycksTest_high_school_microeconomics_5",
        "harness_hendrycksTest_high_school_physics_5",
        "harness_hendrycksTest_high_school_psychology_5",
        "harness_hendrycksTest_high_school_statistics_5",
        "harness_hendrycksTest_high_school_us_history_5",
        "harness_hendrycksTest_high_school_world_history_5",
        "harness_hendrycksTest_human_aging_5",
        "harness_hendrycksTest_human_sexuality_5",
        "harness_hendrycksTest_international_law_5",
        "harness_hendrycksTest_jurisprudence_5",
        "harness_hendrycksTest_logical_fallacies_5",
        "harness_hendrycksTest_machine_learning_5",
        "harness_hendrycksTest_management_5",
        "harness_hendrycksTest_marketing_5",
        "harness_hendrycksTest_medical_genetics_5",
        "harness_hendrycksTest_miscellaneous_5",
        "harness_hendrycksTest_moral_disputes_5",
        "harness_hendrycksTest_moral_scenarios_5",
        "harness_hendrycksTest_nutrition_5",
        "harness_hendrycksTest_philosophy_5",
        "harness_hendrycksTest_prehistory_5",
        "harness_hendrycksTest_professional_accounting_5",
        "harness_hendrycksTest_professional_law_5",
        "harness_hendrycksTest_professional_medicine_5",
        "harness_hendrycksTest_professional_psychology_5",
        "harness_hendrycksTest_public_relations_5",
        "harness_hendrycksTest_security_studies_5",
        "harness_hendrycksTest_sociology_5",
        "harness_hendrycksTest_us_foreign_policy_5",
        "harness_hendrycksTest_virology_5",
        "harness_hendrycksTest_world_religions_5",
    ],
}



import logging
import os
import pickle
from typing import List, Optional

import numpy as np
import requests
from scipy.optimize import minimize

from mergenetic import PROJECT_ROOT

logger = logging.getLogger(__name__)

# ==========================
#  UTILITY FUNCTIONS
# ==========================


def sigmoid(z: np.ndarray) -> np.ndarray:
    """
    Computes the sigmoid function on the given array.

    Parameters
    ----------
    z : np.ndarray
        The input array on which to compute the sigmoid.

    Returns
    -------
    np.ndarray
        The element-wise sigmoid of z.
    """
    return 1 / (1 + np.exp(-z))


def item_curve(theta: np.ndarray, A: np.ndarray, B: np.ndarray) -> np.ndarray:
    """
    Computes the item response curve using a logistic (sigmoid) function,
    given ability parameters theta, discrimination A, and difficulty B.

    Parameters
    ----------
    theta : np.ndarray
        Ability parameter, shaped (1, D, 1) or (N, D, 1).
    A : np.ndarray
        Discrimination parameters, shaped (D, D, N) or (D, D, ...).
    B : np.ndarray
        Difficulty parameters, shaped (D, D, N) or (D, D, ...).

    Returns
    -------
    np.ndarray
        The probability of a correct response under the logistic model.
    """
    # Clip the linear combination to avoid numerical overflow in exp()
    z = np.clip(A * theta - B, -30, 30).sum(axis=1)
    return sigmoid(z)


def fit_theta(
    responses_test: np.ndarray,
    seen_items: List[int],
    A: np.ndarray,
    B: np.ndarray,
    theta_init: Optional[np.ndarray] = None,
    eps: float = 1e-10,
    optimizer: str = "BFGS",
) -> np.ndarray:
    """
    Fits the ability parameter theta by minimizing the negative log-likelihood.

    Parameters
    ----------
    responses_test : np.ndarray
        1D array of observed correct (1) / incorrect (0) responses.
    seen_items : List[int]
        Indices of the items that have been "seen" (answered).
    A : np.ndarray
        Discrimination parameters.
    B : np.ndarray
        Difficulty parameters.
    theta_init : np.ndarray, optional
        Initial guess for theta. Defaults to zeros if not provided.
    eps : float, optional
        A small constant to avoid log(0). Defaults to 1e-10.
    optimizer : str, optional
        The optimization method for scipy.minimize. Defaults to "BFGS".

    Returns
    -------
    np.ndarray
        The fitted theta parameter, shaped (1, D, 1).
    """
    D = A.shape[1]

    def neg_log_like(x):
        P = item_curve(
            x.reshape(1, D, 1), A[:, :, seen_items], B[:, :, seen_items]
        ).squeeze()
        ll = np.sum(
            responses_test[seen_items] * np.log(P + eps)
            + (1 - responses_test[seen_items]) * np.log(1 - P + eps)
        )
        return -ll

    if theta_init is None:
        theta_init = np.zeros(D)

    optimal_theta = minimize(neg_log_like, theta_init, method=optimizer).x
    return optimal_theta[None, :, None]


def fit_lambda(
    responses_test: np.ndarray,
    seen_items: List[int],
    A: np.ndarray,
    B: np.ndarray,
    thetas: List[np.ndarray],
    lambda_init: Optional[np.ndarray] = None,
    eps: float = 1e-10,
    optimizer: str = "BFGS",
) -> np.ndarray:
    """
    Finds the linear combination of multiple theta vectors that best fits
    the observed responses.

    Parameters
    ----------
    responses_test : np.ndarray
        1D array of observed correct (1) / incorrect (0) responses.
    seen_items : List[int]
        Indices of the items that have been "seen" (answered).
    A : np.ndarray
        Discrimination parameters.
    B : np.ndarray
        Difficulty parameters.
    thetas : List[np.ndarray]
        List of candidate theta parameters for linear combination.
    lambda_init : np.ndarray, optional
        Initial guess for the combination weights. Defaults to an even split.
    eps : float, optional
        Small constant for numerical stability. Defaults to 1e-10.
    optimizer : str, optional
        Optimization method. Defaults to "BFGS".

    Returns
    -------
    np.ndarray
        The optimal weights (lambdas) for combining the given thetas.
    """
    D = A.shape[1]

    def neg_log_like(lambdas):
        # Combine thetas with the given lambdas
        combined_theta = sum(l * t for l, t in zip(lambdas, thetas))
        P = item_curve(
            combined_theta.reshape(1, D, 1), A[:, :, seen_items], B[:, :, seen_items]
        ).squeeze()

        log_likelihood = np.sum(
            responses_test[seen_items] * np.log(P + eps)
            + (1 - responses_test[seen_items]) * np.log(1 - P + eps)
        )
        return -log_likelihood

    if lambda_init is None:
        lambda_init = np.ones(len(thetas)) / len(thetas)

    result = minimize(neg_log_like, lambda_init, method=optimizer)
    return result.x


# ==========================
#  TINYBENCHMARKS I/O
# ==========================


def download_tinybenchmarks_if_needed(
    filename: str = f"{PROJECT_ROOT}/tinyBenchmarks.pkl",
) -> None:
    """
    Downloads the tinyBenchmarks.pkl file from the repository if it is not present locally.

    Parameters
    ----------
    filename : str, optional
        Name of the local file to store the data. Defaults to "tinyBenchmarks.pkl".
    """
    if not os.path.isfile(filename):
        url = "https://raw.githubusercontent.com/felipemaiapolo/tinyBenchmarks/main/tinyBenchmarks/tinyBenchmarks.pkl"
        response = requests.get(url)
        if response.status_code == 200:
            with open(filename, "wb") as file:
                file.write(response.content)


def get_tinybenchmarks(filename: str = f"{PROJECT_ROOT}/tinyBenchmarks.pkl") -> dict:
    """
    Downloads the tinyBenchmarks.pkl file from the repository if it is not present locally.

    Parameters
    ----------
    filename : str, optional
        Local name of the tinyBenchmarks file. Defaults to "tinyBenchmarks.pkl".

    Returns
    -------
    dict
        The loaded dictionary containing tinyBenchmarks data.
    """
    download_tinybenchmarks_if_needed(filename)
    with open(filename, "rb") as handle:
        return pickle.load(handle)


# ==========================
#  TOP-LEVEL EVALUATION FUNCTIONS
# ==========================


def evaluate(
    y_input: np.ndarray, bench: str, tinyBenchmarks: Optional[dict] = None
) -> dict:
    """
    Main function that evaluates performance on the chosen benchmark using a combination
    of IRT-based scoring and anchor-based weighting.

    Parameters
    ----------
    y_input : np.ndarray
        A 1D array of binary outcomes (correct=1, incorrect=0) for seen examples.
    bench : str
        The name of the benchmark to evaluate. e.g., "gsm8k".
    tinyBenchmarks : dict, optional
        A loaded tinyBenchmarks dictionary. If None, it will be downloaded or loaded from disk.

    Returns
    -------
    dict
        Dictionary containing the computed estimates (IRT, PIRT, GPIRT) per scenario.
    """
    number_of_examples = 20
    lb_scenarios = ["truthfulqa", "gsm8k", "winogrande", "arc", "hellaswag"]
    benchs = ["lb", "mmlu", "helm_lite", "alpaca"]

    assert len(y_input.shape) == 1, "y_input must be a unidimensional numpy array."
    assert bench in benchs + lb_scenarios, f"Unknown benchmark '{bench}'."

    # If it's a "leaderboard scenario", map bench to "lb"
    bench_name = "lb" if bench in lb_scenarios else bench

    # Load tinyBenchmarks if needed
    if tinyBenchmarks is None:
        tinyBenchmarks = get_tinybenchmarks()

    # Extract relevant data
    data = tinyBenchmarks[bench_name]
    seen_examples = data["seen_examples"]
    examples_weights = data["examples_weights"]
    irt_parameters = data["irt_parameters"]
    A, B = irt_parameters["A"], irt_parameters["B"]
    optimal_lambdas = data["optimal_lambdas"]
    scenarios_position = data["scenarios_position"]
    subscenarios_position = data["subscenarios_position"]

    N = max(np.max(x) for x in scenarios_position.values()) + 1

    # Precompute scenario-based balancing
    balance_weights = np.ones(N)
    for scenario in scenarios_position.keys():
        N_sce = len(scenarios_position[scenario])
        n_sub = len(subscenarios_position[scenario])
        for sub in subscenarios_position[scenario].keys():
            n_i = len(subscenarios_position[scenario][sub])
            balance_weights[subscenarios_position[scenario][sub]] = N_sce / (
                n_sub * n_i
            )

    # If using the big IRT model for a single scenario
    if bench not in benchs:
        scenario_idx = [
            i for i, s in enumerate(scenarios_position.keys()) if s == bench
        ][0]
        start = number_of_examples * scenario_idx
        seen_examples = seen_examples[start : start + number_of_examples]
    else:
        # Evaluate all scenarios in "lb", "mmlu", etc.
        pass

    # Create y vector from the input data
    y = np.zeros(N)
    for i, ex_id in enumerate(seen_examples):
        y[ex_id] = y_input[i]

    # Fit user ability parameter
    theta = fit_theta(y, seen_examples, A, B)

    # Compute estimates for each scenario
    estimates = {}
    unseen_examples = [i for i in range(N) if i not in seen_examples]

    for scenario in scenarios_position.keys():
        # Scenario info
        N_sce = len(scenarios_position[scenario])
        seen_sce = [s for s in seen_examples if s in scenarios_position[scenario]]
        unseen_sce = [s for s in unseen_examples if s in scenarios_position[scenario]]

        data_part_IRTp = ((balance_weights * y)[seen_sce]).mean() if seen_sce else 0.0
        irt_part = (
            (balance_weights * item_curve(theta.reshape(1, A.shape[1], 1), A, B))[
                0, unseen_sce
            ].mean()
            if unseen_sce
            else 0.0
        )

        # Weighted blending
        IRTp_lambd = number_of_examples / N_sce
        IRTp = IRTp_lambd * data_part_IRTp + (1 - IRTp_lambd) * irt_part

        # Weighted sum of correct answers
        IRT = (examples_weights[scenario] * y[seen_sce]).sum() if seen_sce else 0.0

        # Combine with the scenario's optimal lambda
        IRTpp = optimal_lambdas[scenario] * IRT + (1 - optimal_lambdas[scenario]) * IRTp

        estimates[scenario] = {"irt": IRT, "pirt": IRTp, "gpirt": IRTpp}

    return estimates


def estimate_theta(
    y_input: np.ndarray,
    bench: str,
    number_of_examples: int = 100,
    tinyBenchmarks: Optional[dict] = None,
) -> np.ndarray:
    """
    Estimates the ability parameter theta for the chosen scenario in the LB set.

    Parameters
    ----------
    y_input : np.ndarray
        1D array of binary outcomes.
    bench : str
        The scenario in lb_scenarios, e.g., "gsm8k".
    number_of_examples : int, optional
        Number of examples used for the scenario. Defaults to 100.
    tinyBenchmarks : dict, optional
        Data dictionary. If None, it is loaded from disk or downloaded.

    Returns
    -------
    np.ndarray
        The estimated theta parameter, shaped (1, D, 1).
    """
    lb_scenarios = ["truthfulqa", "gsm8k", "winogrande", "arc", "hellaswag"]

    assert len(y_input.shape) == 1, "y_input must be unidimensional."
    assert bench in lb_scenarios, f"Benchmark '{bench}' not recognized."
    bench_name = "lb"  # For LB-based scenarios

    if tinyBenchmarks is None:
        tinyBenchmarks = get_tinybenchmarks()
        logger.info("Loaded official tinyBenchmarks in estimate_theta...")

    data = tinyBenchmarks[bench_name]
    A, B = data["irt_parameters"]["A"], data["irt_parameters"]["B"]
    seen_examples = data["seen_examples"]
    scenarios_position = data["scenarios_position"]

    # Identify the relevant scenario block
    scenario_idx = [i for i, s in enumerate(scenarios_position.keys()) if s == bench][0]
    start = number_of_examples * scenario_idx
    scenario_seen_examples = seen_examples[start : start + number_of_examples]

    # Shift to the scenario's index range
    shift = scenarios_position[bench][0]
    scenario_seen_examples = [ex + shift for ex in scenario_seen_examples]

    # Construct the y vector
    N = max(np.max(x) for x in scenarios_position.values()) + 1
    y = np.zeros(N)
    for i, ex_id in enumerate(scenario_seen_examples):
        y[ex_id] = y_input[i]

    return fit_theta(y, scenario_seen_examples, A, B)


def estimate_theta_linear(
    y_input: np.ndarray,
    bench: str,
    thetas: List[np.ndarray],
    seen_examples: List[int],
    tinyBenchmarks: Optional[dict] = None,
) -> np.ndarray:
    """
    Estimates a combined theta by finding the optimal linear combination of thetas
    that best fits the observed data.

    Parameters
    ----------
    y_input : np.ndarray
        1D array of binary outcomes (correct=1, incorrect=0).
    bench : str
        A scenario from the LB set (e.g., "gsm8k").
    thetas : List[np.ndarray]
        Candidate theta parameters.
    seen_examples : List[int]
        Indices of items that have been answered (seen).
    tinyBenchmarks : dict, optional
        Data dictionary. If None, it will be loaded from disk.

    Returns
    -------
    np.ndarray
        The combined theta array, shaped (1, D, 1).
    """
    lb_scenarios = ["truthfulqa", "gsm8k", "winogrande", "arc", "hellaswag"]

    assert len(y_input.shape) == 1, "y_input must be unidimensional."
    assert bench in lb_scenarios, f"Benchmark '{bench}' not recognized."
    bench_name = "lb"

    if tinyBenchmarks is None:
        tinyBenchmarks = get_tinybenchmarks()

    data = tinyBenchmarks[bench_name]
    A, B = data["irt_parameters"]["A"], data["irt_parameters"]["B"]
    scenarios_position = data["scenarios_position"]

    # Construct the y vector
    N = max(np.max(x) for x in scenarios_position.values()) + 1
    y = np.zeros(N)
    for i, ex_id in enumerate(seen_examples):
        y[ex_id] = y_input[i]

    # Fit the lambda weights
    lambdas_opt = fit_lambda(y, seen_examples, A, B, thetas)
    # Weighted sum of the candidate thetas
    combined_theta = sum(l * t for l, t in zip(lambdas_opt, thetas))

    return combined_theta


def estimate_theta_anchors(
    y_input: np.ndarray,
    bench: str,
    anchor_ids: List[int],
    tinyBenchmarks: Optional[dict] = None,
) -> np.ndarray:
    """
    Estimate the ability parameter (theta) using only the specified anchor IDs.

    Parameters
    ----------
    y_input : np.ndarray
        1D array of binary outcomes for anchor items.
    bench : str
        A scenario from the LB set (e.g., "gsm8k").
    anchor_ids : List[int]
        The local item indices used as anchors for fitting.
    tinyBenchmarks : dict, optional
        Data dictionary. If None, it will be loaded from disk.

    Returns
    -------
    np.ndarray
        The estimated theta array, shaped (1, D, 1).
    """
    lb_scenarios = ["truthfulqa", "gsm8k", "winogrande", "arc", "hellaswag"]
    assert len(y_input.shape) == 1, "y_input must be unidimensional."
    assert bench in lb_scenarios, f"Benchmark '{bench}' not recognized."

    bench_name = "lb"
    if tinyBenchmarks is None:
        tinyBenchmarks = get_tinybenchmarks()

    data = tinyBenchmarks[bench_name]
    A, B = data["irt_parameters"]["A"], data["irt_parameters"]["B"]
    scenarios_position = data["scenarios_position"]

    shift = np.array(scenarios_position[bench]).min()
    seen_examples = list(np.array(anchor_ids) + shift)

    N = max(np.max(x) for x in scenarios_position.values()) + 1
    y = np.zeros(N)
    for i, ex_id in enumerate(seen_examples):
        y[ex_id] = y_input[i]

    return fit_theta(y, seen_examples, A, B)


def eval_mpirt_on_anchors(
    y_input: np.ndarray,
    thetas: List[np.ndarray],
    bench: str,
    anchors_idx: List[int],
    tinyBenchmarks: dict = None,
    examples_weights: Optional[np.ndarray] = None,
) -> float:
    """
    Evaluate merged-performance IRT on anchor items, combining candidate thetas linearly.

    Parameters
    ----------
    y_input : np.ndarray
        1D array of binary outcomes on anchors.
    thetas : List[np.ndarray]
        Candidate thetas to combine.
    bench : str
        The scenario name (from lb_scenarios).
    anchors_idx : List[int]
        Local indices of anchor items.
    tinyBenchmarks : dict
        The loaded tinyBenchmarks data.
    examples_weights : np.ndarray, optional
        Weights for the anchor items. Defaults to None.

    Returns
    -------
    float
        The partial IRT estimate.
    """
    num_of_examples = y_input.shape[0]
    SCENARIOS = ["truthfulqa", "gsm8k", "winogrande", "arc", "hellaswag"]

    assert len(y_input.shape) == 1, "y_input must be a unidimensional numpy array."
    assert bench in SCENARIOS, f"Unknown scenario '{bench}'"

    if tinyBenchmarks is None:
        tinyBenchmarks = get_tinybenchmarks()

    bench_name = "lb"
    irt_params = tinyBenchmarks[bench_name]["irt_parameters"]
    A, B = irt_params["A"], irt_params["B"]
    scenarios_position = tinyBenchmarks[bench_name]["scenarios_position"]

    shift = scenarios_position[bench][0]
    seen_examples = [idx + shift for idx in anchors_idx]

    if examples_weights is None:
        # Possibly rely on some default weighting or user-supplied weights
        examples_weights = np.ones_like(y_input) / num_of_examples

    # Construct full y
    N = max(np.max(x) for x in scenarios_position.values()) + 1
    y = np.zeros(N)
    for i, ex_id in enumerate(seen_examples):
        y[ex_id] = y_input[i]

    # Combine or pick thetas
    if len(thetas) == 1:
        final_theta = thetas[0]
    else:
        final_theta = estimate_theta_linear(
            y_input, bench, thetas, seen_examples, tinyBenchmarks
        )

    # Identify seen/unseen for scenario
    unseen_examples = [i for i in range(N) if i not in seen_examples]
    unseen_sce = [s for s in unseen_examples if s in scenarios_position[bench]]
    # Weighted item-curve
    if len(unseen_sce) > 0:
        mpirt_part = (item_curve(final_theta.reshape(1, A.shape[1], 1), A, B))[
            0, unseen_sce
        ].mean()
    else:
        mpirt_part = 0.0

    # Weighted "observed" portion
    seen_sce = [s for s in seen_examples if s in scenarios_position[bench]]
    w_irt = (examples_weights * y_input).sum() if len(seen_sce) > 0 else 0.0

    # Weighted combination
    lambd = num_of_examples / len(scenarios_position[bench])
    result = lambd * w_irt + (1 - lambd) * mpirt_part
    return result


def estimate_fitness(
    y_input: np.ndarray,
    thetas: List[np.ndarray],
    bench: str,
    anchors_idx: List[int],
    example_weights: np.ndarray,
    tinybenchmarks: dict = None,
    delta: float = 0.5,
) -> dict:
    """
    Evaluate merge performance on anchors, returning mpirt, rc, and a combination gmpirt.

    Parameters
    ----------
    y_input : np.ndarray
        1D array of binary outcomes for anchor items.
    thetas : List[np.ndarray]
        Candidate thetas for combination.
    bench : str
        Scenario name from lb_scenarios.
    anchors_idx : List[int]
        Local indices of anchor items.
    example_weights : np.ndarray
        Weights for the anchor items.
    tinybenchmarks : dict
        The loaded tinyBenchmarks data.
    delta : float, optional
        Blend factor for combining anchor-based correctness and mpirt. Defaults to 0.5.

    Returns
    -------
    dict
        Contains "mpirt", "weighted_avg", and "gmpirt" measures.
    """
    mpirt_val = eval_mpirt_on_anchors(
        y_input,
        thetas,
        bench,
        anchors_idx,
        tinybenchmarks,
        examples_weights=example_weights,
    )
    weighted_avg = (example_weights * y_input).sum()
    gmpirt = delta * weighted_avg + (1 - delta) * mpirt_val

    names = ["mpirt", "gmpirt"] if len(thetas) > 1 else ["pirt", "gpirt"]

    return {names[0]: mpirt_val, "weighted_avg": weighted_avg, names[1]: gmpirt}



#!/usr/bin/env python

import os

import setuptools

# --- CUDA Extension Setup ---
# This list will hold your Extension objects
ext_modules = []

# Environment variable to explicitly skip CUDA build
skip_cuda_build = os.environ.get("SKIP_CUDA_BUILD", "0") == "1"

if skip_cuda_build:
    print("---------------------------------------------------------------------")
    print("SKIP_CUDA_BUILD is set to 1. Skipping CUDA extension compilation.")
    print("---------------------------------------------------------------------")
else:
    # Attempt to build CUDA extensions if not explicitly skipped
    cuda_home = os.environ.get("CUDA_HOME")
    # You might want a more robust check for nvcc or other CUDA components
    if cuda_home:
        print(f"---------------------------------------------------------------------")
        print(f"CUDA_HOME found at {cuda_home}. Attempting to build CUDA extensions.")
        print(
            f"If you want to skip this, set the SKIP_CUDA_BUILD=1 environment variable."
        )
        print(f"---------------------------------------------------------------------")

        # ==========================================================================
        # TODO: DEFINE YOUR CUDA EXTENSIONS HERE
        # This is where you would define your setuptools.Extension objects for
        # your CUDA code.
        #
        # Example (replace with your actual extension details):
        #
        # my_cuda_extension = Extension(
        #     name="mergenetic.cuda_ops",  # How it will be imported, e.g., from mergenetic import cuda_ops
        #     sources=[
        #         "src/mergenetic/cuda_ops/ops_interface.cpp", # C++/Pybind11 wrapper
        #         "src/mergenetic/cuda_ops/kernels.cu",      # Your CUDA kernel file
        #     ],
        #     include_dirs=[
        #         os.path.join(cuda_home, "include"),
        #         # any other include directories for your C++/CUDA code
        #     ],
        #     library_dirs=[os.path.join(cuda_home, "lib64")], # Adjust lib/lib64 as per your CUDA setup
        #     libraries=["cudart"], # CUDA runtime library, add others if needed (e.g., cublas, cufft)
        #     runtime_library_dirs=[os.path.join(cuda_home, "lib64")],
        #     # Extra compile args for nvcc (via C++ compiler) or direct C++ flags
        #     extra_compile_args={
        #         'nvcc': [ # For .cu files if setuptools uses NVCC directly or via a custom build step
        #             '-O3',
        #             # Add other NVCC flags, e.g., -gencode arch=compute_XX,code=sm_XX
        #         ],
        #         'cxx': [ # For .cpp files
        #             '-O3',
        #             # Add other C++ compiler flags
        #         ]
        #     },
        #     language="c++" # Or "cuda" if your setuptools/build system handles it directly
        # )
        # ext_modules.append(my_cuda_extension)
        #
        # If you have multiple CUDA extensions, define and append each one.
        # ==========================================================================

        if (
            not ext_modules
        ):  # If no extensions were defined (e.g., placeholder not filled)
            print(
                "---------------------------------------------------------------------"
            )
            print(
                "WARNING: CUDA detected, but no CUDA extensions were defined in setup.py."
            )
            print("The package will be built without CUDA-specific compiled modules.")
            print(
                "Please fill in the 'TODO' section in setup.py if CUDA modules are expected."
            )
            print(
                "---------------------------------------------------------------------"
            )

    else:
        print("---------------------------------------------------------------------")
        print("CUDA_HOME environment variable not found.")
        print("Skipping CUDA extension compilation.")
        print(
            "If you have CUDA installed and want to build extensions, ensure CUDA_HOME is set."
        )
        print(
            "Alternatively, set SKIP_CUDA_BUILD=1 to suppress this message and build without CUDA."
        )
        print("---------------------------------------------------------------------")

if __name__ == "__main__":
    setuptools.setup(
        ext_modules=ext_modules  # Add the (possibly empty) list of extensions here
    )



import os
import pickle
from dataclasses import dataclass
from enum import Enum
from logging import getLogger
from pathlib import Path

import pandas as pd
import torch
from lm_eval.models.vllm_causallms import VLLM

from mergenetic.estimator.utils import estimate_theta_anchors
from mergenetic.evaluation import BaseEvaluator, FGMathEvaluator, MCEvaluator
from mergenetic.evaluation.lm_harness import LmHarnessEvaluator
from mergenetic.utils import (
    Config,
    ConfigLmEval,
    get_batched_model_predictions,
    loading_model_tokenizer_gpu_only_4bit_local,
)

logger = getLogger(__name__)

############################################
# General utility functions                #
############################################


def check_file_exists(file_path: str) -> bool:
    """
    Check if the file exists at the given path
    """
    return Path(file_path).exists()


############################################
# Configuration utility functions          #
############################################


@dataclass
class ConfigEvaluation:
    """
    Configuration Required to request the evaluation for either a mathematical or multiple choice task.
    The dataset, the model, and an output where to save the file must be furnished.
    The file is just a csv with the predictions that can be used to compute the score.
    """

    eval_task: str = ""  # MATH or MULTIPLECHOICE
    dataset_path: str = ""
    model_path: str = ""
    output_path: str = ""
    dtype: str = ""
    eval_batch_size: int = 8


@dataclass
class ConfigThetaEstimation:
    """
    A configuration class to request the estimation of the latent ability of the model.
    Such latent ability are the thetas.
    Tb_path is needed if you wish to use original tinybenchmarks parameters.
    """

    responses_path: str
    bench: str
    output_path: str


@dataclass
class EvalMode(str, Enum):
    RANDOM = "random"
    MPIRT = "mpirt"
    GMPIRT = "gmpirt"
    RC = "rc"
    ADVERSARIAL = "adversarial"
    IDEAL = "ideal"


def validate_evaluation_choice(choice):
    """
    Check if the passed evaluation choice is valid
    """
    try:
        EvalMode(choice)
    except ValueError:
        raise ValueError(
            "Invalid evaluation mode. Choose 'random', 'mpirt', 'gmpirt', or 'rc'."
        )


############################################
# Evaluation functions
############################################
class Tasks(Enum):
    FG_MATH = "FG_MATH"
    MULTIPLECHOICE = "MULTIPLECHOICE"


def task_2_fitness(task: str) -> BaseEvaluator:
    """
    Map the task to the evaluator
    """
    if task == Tasks.FG_MATH.value:
        return FGMathEvaluator
    elif task == Tasks.MULTIPLECHOICE.value:
        return MCEvaluator
    else:
        raise ValueError("Invalid task. Choose 'FG_MATH' or 'MULTIPLECHOICE'.")


def evaluate(
    task: str,
    df: pd.DataFrame,
    model_path: str,
    eval_batch_size: int,
    load_on_gpu: bool,
    get_also_series: bool = True,
) -> float | tuple[float, pd.DataFrame]:

    assert (
        task in Tasks.__members__.keys()
    ), f"Invalid task '{task}'. Choose {Tasks.__members__.keys()}"

    # load the model and tokenizer
    model, tokenizer = loading_model_tokenizer_gpu_only_4bit_local(model_path)

    # get the predictions
    if load_on_gpu:
        device = "cuda:0"
    else:
        device = "cpu"
    df["predictions"] = get_batched_model_predictions(
        model,
        tokenizer,
        df,
        batch_size=eval_batch_size,
        randomness=False,
        device=device,
        print_output=True,
        apply_chat=False,
    )

    # get the correctness
    evaluator: BaseEvaluator = task_2_fitness(task)()
    f = -1 * evaluator.get_correctness(df)

    if get_also_series:
        return [f], evaluator.get_data()
    else:
        return [f]


def evaluate_models(config: Config):
    """
    Evaluate the model on the given task
    """
    if config.datasets:
        # zip together the model paths and output paths
        models = dict(config.models.items())

        langs = config.datasets.keys()
        output_paths = {
            k: config.path_to_store_config + "/" + model.split("/")[-1] + f"_{k}.csv"
            for k, model in config.models.items()
        }

        if config.base_model is not None:
            models["base"] = config.base_model
            output_paths["base"] = (
                config.path_to_store_config
                + "/"
                + config.base_model.split("/")[-1]
                + "_base.csv"
            )
            langs = list(langs) + ["base"]

        evaluations = {}
        for lang in langs:
            model_path = models[lang]
            output_path = output_paths[lang]

            # check if file exists
            if not config.force_evaluation and check_file_exists(output_path):
                logger.info(
                    f"Skipping evaluation for {model_path} as the evaluation file already exists."
                )
                # load the file
                df = pd.read_csv(output_path)
                evaluations[lang] = df
                continue

            logger.info(
                f"Evaluation for {model_path} not found at path {output_path}. Starting evaluation."
            )

            df = pd.read_csv(config.datasets[lang])
            if model_path != "":
                logger.info(f"Head of the dataframe for {lang}: {df.head()}")
                fit, pred_df = evaluate(
                    config.task_type, df, model_path, config.eval_batch_size, True
                )
                if output_path != "":
                    pred_df.to_csv(output_path)
                    logger.info(
                        f"The predictions for {lang} are saved to {output_path}. Fitness: {fit}"
                    )
                evaluations[lang] = pred_df
    else:
        raise ValueError("No datasets provided for evaluation.")

    return evaluations


def evaluate_models_lm_harness(config: ConfigLmEval):
    """
    Evaluate the model on the given search tasks
    """
    tasks = config.tasks["search"]
    langs = config.langs
    output_paths = [
        config.path_to_store_config + "/lm-eval_" + model.split("/")[-1] + f"_{k}.csv"
        for k, model in config.models.items()
    ]
    model_output_langs = zip(output_paths, langs)
    evaluations = {}

    for output_path, lang in model_output_langs:
        # check if file exists
        task = tasks[lang]
        if not config.force_evaluation and check_file_exists(output_path):
            logger.info(
                f"Skipping evaluation for {lang} as the evaluation file already exists."
            )
            # load the file
            df = pd.read_csv(output_path)
            evaluations[lang] = df
            continue

        logger.info(
            f"Evaluation for {lang} not found at path {output_path}. Starting evaluation."
        )

        model_path = config.models[lang]
        model = VLLM(
            pretrained=str(model_path),
            device=config.device,
            dtype=torch.bfloat16,
            quantization="bitsandbytes" if config.load_in_4bit else None,
            gpu_memory_utilization=0.8 if config.load_in_4bit else 0.9,
        )

        evaluator = LmHarnessEvaluator(
            task_name=task,
            sample_ids=None,
            correctness_metric=config.metric,
            lang_id=None,
            is_test=False,
            additional_templates_folder=config.additional_templates_folder,
            batch_size=config.eval_batch_size,
        )
        evaluator.evaluate(model)
        evaluations[lang] = evaluator.get_data()
        evaluations[lang].to_csv(output_path)
        logger.info(f"The predictions for {lang} are saved to {output_path}.")

    if config.base_model is not None:
        out_path = f"{config.path_to_store_config}/lm-eval_{config.base_model.split('/')[-1]}_base.csv"

        if not config.force_evaluation and check_file_exists(out_path):
            logger.info(
                f"Skipping evaluation for the base model as the evaluation file already exists."
            )
            # load the file
            evaluations["base"] = pd.read_csv(out_path)
            return evaluations

        logger.info(
            f"Evaluation for the base model not found at path {out_path}. Starting evaluation."
        )
        base_model = VLLM(
            pretrained=str(config.base_model),
            device=config.device,
            quantization="bitsandbytes" if config.load_in_4bit else None,
            gpu_memory_utilization=0.8 if config.load_in_4bit else 0.9,
        )
        evaluator = LmHarnessEvaluator(
            task_name=tasks["base"],
            sample_ids=None,
            correctness_metric=config.metric,
            lang_id=None,
            is_test=True,
            additional_templates_folder=config.additional_templates_folder,
            batch_size=config.eval_batch_size,
        )
        evaluator.evaluate(base_model)
        evaluations["base"] = evaluator.get_data()
        evaluations["base"].to_csv(out_path)


############################################
# Theta Estimation functions
############################################


def extract_thetas(config: Config):
    """
    Extract thetas for models in ft_models
    """
    responses_paths = {
        l: f"{config.path_to_store_config}/{m.split('/')[-1]}_{l}.csv"
        for l, m in config.models.items()
    }
    output_path_theta_estimation = {
        l: f"{config.path_to_store_config}/{m.split('/')[-1]}_{l}_theta.pkl"
        for l, m in config.models.items()
    }

    if config.base_model is not None:
        responses_paths["base"] = (
            f"{config.path_to_store_config}/{config.base_model.split('/')[-1]}_base.csv"
        )
        output_path_theta_estimation["base"] = (
            f"{config.path_to_store_config}/{config.base_model.split('/')[-1]}_base_theta.pkl"
        )

    thetas = {}

    for l in responses_paths.keys():
        responses_path = responses_paths[l]
        output_path = output_path_theta_estimation[l]

        # init config
        config_theta = ConfigThetaEstimation(
            responses_path=responses_path,
            bench=config.bench,
            output_path=output_path,
        )

        try:
            responses = pd.read_csv(config_theta.responses_path)
        except FileNotFoundError:
            logger.info(f"Responses file not found at {config_theta.responses_path}.")

            logger.info(f"Looking for lm-evaluation-harness evaluation files.")
            model_nm = config.models[l] if l != "base" else config.base_model
            responses_path = f"{config.path_to_store_config}/lm-eval_{model_nm.split('/')[-1]}_{l}.csv"
            responses = pd.read_csv(responses_path)
        y = responses["correctness"].values
        anchors = list(range(len(y)))

        theta = estimate_theta_anchors(y, config_theta.bench, anchors)
        if config_theta.output_path != "":
            if not os.path.exists(config_theta.output_path):
                os.makedirs(os.path.dirname(config_theta.output_path), exist_ok=True)
            with open(config_theta.output_path, "wb") as f:
                pickle.dump(theta, f)
                logger.info(f"Theta saved to {config_theta.output_path}")
        thetas[l] = theta

    return thetas


def retrieve_thetas(config: Config):
    """
    Retrieve thetas from the given path
    """
    thetas_paths = {
        k: f"{config.path_to_store_config}/{m.split('/')[-1]}_{k}_theta.pkl"
        for k, m in config.models.items()
    }

    if config.base_model is not None:
        thetas_paths["base"] = (
            f"{config.path_to_store_config}/{config.base_model.split('/')[-1]}_base_theta.pkl"
        )

    thetas = {}
    for k, path in thetas_paths.items():
        if check_file_exists(path):
            with open(path, "rb") as f:
                thetas[k] = pickle.load(f)
            logger.info(f"Theta for '{k}' retrieved from {path}.")
        else:
            logger.info(f"Theta for '{k}' not found at {path}. Extracting thetas.")
            thetas.update(extract_thetas(config))
            break

    return thetas
